import logging
logger = logging.getLogger(__name__)


class MetaDriveType:
    """
    Following waymo style, this class defines a set of strings used to denote different types of objects.
    Those types are used within MetaDrive and might mismatch to the strings used in other dataset.

    NOTE: when add new keys, make sure class method works well for them
    """

    # ===== Lane, Road =====
    LANE_SURFACE_STREET = "LANE_SURFACE_STREET"
    LANE_UNKNOWN = "LANE_UNKNOWN"
    LANE_FREEWAY = "LANE_FREEWAY"
    LANE_BIKE_LANE = "LANE_BIKE_LANE"

    # ===== Lane Line =====
    LINE_UNKNOWN = "UNKNOWN_LINE"
    LINE_BROKEN_SINGLE_WHITE = "ROAD_LINE_BROKEN_SINGLE_WHITE"
    LINE_SOLID_SINGLE_WHITE = "ROAD_LINE_SOLID_SINGLE_WHITE"
    LINE_SOLID_DOUBLE_WHITE = "ROAD_LINE_SOLID_DOUBLE_WHITE"
    LINE_BROKEN_SINGLE_YELLOW = "ROAD_LINE_BROKEN_SINGLE_YELLOW"
    LINE_BROKEN_DOUBLE_YELLOW = "ROAD_LINE_BROKEN_DOUBLE_YELLOW"
    LINE_SOLID_SINGLE_YELLOW = "ROAD_LINE_SOLID_SINGLE_YELLOW"
    LINE_SOLID_DOUBLE_YELLOW = "ROAD_LINE_SOLID_DOUBLE_YELLOW"
    LINE_PASSING_DOUBLE_YELLOW = "ROAD_LINE_PASSING_DOUBLE_YELLOW"

    # ===== Edge/Boundary/SideWalk/Region =====
    BOUNDARY_UNKNOWN = "UNKNOWN"
    BOUNDARY_LINE = "ROAD_EDGE_BOUNDARY"
    BOUNDARY_MEDIAN = "ROAD_EDGE_MEDIAN"
    STOP_SIGN = "STOP_SIGN"
    CROSSWALK = "CROSSWALK"
    SPEED_BUMP = "SPEED_BUMP"
    DRIVEWAY = "DRIVEWAY"

    # ===== Traffic Light =====
    LANE_STATE_UNKNOWN = "LANE_STATE_UNKNOWN"
    LANE_STATE_ARROW_STOP = "LANE_STATE_ARROW_STOP"
    LANE_STATE_ARROW_CAUTION = "LANE_STATE_ARROW_CAUTION"
    LANE_STATE_ARROW_GO = "LANE_STATE_ARROW_GO"
    LANE_STATE_STOP = "LANE_STATE_STOP"
    LANE_STATE_CAUTION = "LANE_STATE_CAUTION"
    LANE_STATE_GO = "LANE_STATE_GO"
    LANE_STATE_FLASHING_STOP = "LANE_STATE_FLASHING_STOP"
    LANE_STATE_FLASHING_CAUTION = "LANE_STATE_FLASHING_CAUTION"

    # the light states above will be converted to the following 4 types
    LIGHT_GREEN = "TRAFFIC_LIGHT_GREEN"
    LIGHT_RED = "TRAFFIC_LIGHT_RED"
    LIGHT_YELLOW = "TRAFFIC_LIGHT_YELLOW"
    LIGHT_UNKNOWN = "TRAFFIC_LIGHT_UNKNOWN"

    # ===== Agent type =====
    UNSET = "UNSET"
    VEHICLE = "VEHICLE"
    PEDESTRIAN = "PEDESTRIAN"
    CYCLIST = "CYCLIST"
    OTHER = "OTHER"

    # ===== Object type =====
    TRAFFIC_LIGHT = "TRAFFIC_LIGHT"
    TRAFFIC_BARRIER = "TRAFFIC_BARRIER"
    TRAFFIC_CONE = "TRAFFIC_CONE"
    TRAFFIC_OBJECT = "TRAFFIC_OBJECT"
    GROUND = "GROUND"
    INVISIBLE_WALL = "INVISIBLE_WALL"
    BUILDING = "BUILDING"

    # ===== Coordinate system =====
    COORDINATE_METADRIVE = "metadrive"
    COORDINATE_WAYMO = "waymo"

    # deprecated
    # LIGHT_ENUM_TO_STR = {
    #     0: LANE_STATE_UNKNOWN,
    #     1: LANE_STATE_ARROW_STOP,
    #     2: LANE_STATE_ARROW_CAUTION,
    #     3: LANE_STATE_ARROW_GO,
    #     4: LANE_STATE_STOP,
    #     5: LANE_STATE_CAUTION,
    #     6: LANE_STATE_GO,
    #     7: LANE_STATE_FLASHING_STOP,
    #     8: LANE_STATE_FLASHING_CAUTION
    # }

    @classmethod
    def is_traffic_object(cls, type):
        return type in [cls.TRAFFIC_CONE, cls.TRAFFIC_BARRIER, cls.TRAFFIC_OBJECT]

    @classmethod
    def has_type(cls, type_string: str):
        return type_string in cls.__dict__

    @classmethod
    def from_waymo(cls, waymo_type_string: str):
        assert cls.__dict__[waymo_type_string]
        return waymo_type_string

    @classmethod
    def from_nuplan(cls, waymo_type_string: str):
        # TODO: WIP
        return ""

    @classmethod
    def is_lane(cls, type):
        return type in [cls.LANE_SURFACE_STREET, cls.LANE_FREEWAY, cls.LANE_BIKE_LANE]

    @classmethod
    def is_road_line(cls, line):
        """
        This function relates to is_road_edge. We will have different processing when treating a line that
        is in the boundary or not.
        """
        return line in [
            cls.LINE_UNKNOWN, cls.LINE_BROKEN_SINGLE_WHITE, cls.LINE_SOLID_SINGLE_WHITE, cls.LINE_SOLID_DOUBLE_WHITE,
            cls.LINE_BROKEN_SINGLE_YELLOW, cls.LINE_BROKEN_DOUBLE_YELLOW, cls.LINE_SOLID_SINGLE_YELLOW,
            cls.LINE_SOLID_DOUBLE_YELLOW, cls.LINE_PASSING_DOUBLE_YELLOW
        ]

    @classmethod
    def is_yellow_line(cls, line):
        return line in [
            cls.LINE_SOLID_DOUBLE_YELLOW, cls.LINE_PASSING_DOUBLE_YELLOW, cls.LINE_SOLID_SINGLE_YELLOW,
            cls.LINE_BROKEN_DOUBLE_YELLOW, cls.LINE_BROKEN_SINGLE_YELLOW
        ]

    @classmethod
    def is_white_line(cls, line):
        return MetaDriveType.is_road_line(line) and not MetaDriveType.is_yellow_line(line)

    @classmethod
    def is_broken_line(cls, line):
        return line in [cls.LINE_BROKEN_DOUBLE_YELLOW, cls.LINE_BROKEN_SINGLE_YELLOW, cls.LINE_BROKEN_SINGLE_WHITE]

    @classmethod
    def is_road_edge(cls, edge):
        """
        This function relates to is_road_line.
        """
        return edge in [cls.BOUNDARY_UNKNOWN, cls.BOUNDARY_LINE, cls.BOUNDARY_MEDIAN]

    @classmethod
    def is_sidewalk(cls, edge):
        return edge == cls.BOUNDARY_LINE

    @classmethod
    def is_vehicle(cls, type):
        return type == cls.VEHICLE

    @classmethod
    def is_traffic_light_in_yellow(cls, light):
        return cls.simplify_light_status(light) == cls.LIGHT_YELLOW

    @classmethod
    def is_traffic_light_in_green(cls, light):
        return cls.simplify_light_status(light) == cls.LIGHT_GREEN

    @classmethod
    def is_traffic_light_in_red(cls, light):
        return cls.simplify_light_status(light) == cls.LIGHT_RED

    @classmethod
    def is_traffic_light_unknown(cls, light):
        return cls.simplify_light_status(light) == cls.LIGHT_UNKNOWN

    @classmethod
    def parse_light_status(cls, status: str, simplifying=True):
        """
        Parse light status from ENUM to STR
        """
        # if data_source == "waymo":
        #     status = cls.LIGHT_ENUM_TO_STR[status]
        if simplifying:
            return cls.simplify_light_status(status)
        else:
            return status

    @classmethod
    def simplify_light_status(cls, status: str):
        """
        Convert status to red/yellow/green/unknown
        """
        if status in [cls.LANE_STATE_UNKNOWN, cls.LANE_STATE_FLASHING_STOP, cls.LIGHT_UNKNOWN, None]:
            return cls.LIGHT_UNKNOWN
        elif status in [cls.LANE_STATE_ARROW_STOP, cls.LANE_STATE_STOP, cls.LIGHT_RED]:
            return cls.LIGHT_RED
        elif status in [cls.LANE_STATE_ARROW_CAUTION, cls.LANE_STATE_CAUTION, cls.LANE_STATE_FLASHING_CAUTION,
                        cls.LIGHT_YELLOW]:
            return cls.LIGHT_YELLOW
        elif status in [cls.LANE_STATE_ARROW_GO, cls.LANE_STATE_GO, cls.LIGHT_GREEN]:
            return cls.LIGHT_GREEN
        else:
            logger.warning("TrafficLightStatus: {} is not MetaDriveType".format(status))
            return cls.LIGHT_UNKNOWN



import time

from metadrive.envs.top_down_env import TopDownSingleFrameMetaDriveEnv
from metadrive.utils import setup_logger


def vis_top_down_render_with_panda_render():
    setup_logger(True)

    env = TopDownSingleFrameMetaDriveEnv(
        {
            "num_scenarios": 1,
            "manual_control": True,
            "use_render": True,
            "image_observation": False,
            "traffic_mode": "respawn"
        }
    )
    o, _ = env.reset()
    s = time.time()
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step(env.action_space.sample())
        env.render(
            text={
                "vehicle_num": len(env.engine.traffic_manager.vehicles),
                "traffic_vehicle": len(env.engine.traffic_manager.traffic_vehicles)
            }
        )
        # if d:
        #     env.reset()
        # if i % 1000 == 0:
        # print("Steps: {}, Time: {}".format(i, time.time() - s))
    env.close()


if __name__ == '__main__':
    vis_top_down_render_with_panda_render()



import cv2

from metadrive.component.map.scenario_map import ScenarioMap
from metadrive.engine.asset_loader import AssetLoader
from metadrive.engine.engine_utils import initialize_engine, close_engine
from metadrive.envs.scenario_env import ScenarioEnv
from metadrive.manager.scenario_data_manager import ScenarioDataManager


# @pytest.mark.parametrize("dir", ["waymo"])
def test_map_get_semantic_map(dir="waymo", render=False, show=False):
    default_config = ScenarioEnv.default_config()
    default_config["use_render"] = render
    default_config["debug"] = False
    default_config["debug_static_world"] = False
    default_config["data_directory"] = AssetLoader.file_path(dir, return_raw_style=False)
    # default_config["data_directory"] = AssetLoader.file_path("nuscenes", return_raw_style=False)
    # default_config["data_directory"] = "/home/shady/Downloads/test_processed"
    default_config["num_scenarios"] = 3
    engine = initialize_engine(default_config)
    try:
        size, res = 512, 2
        engine.data_manager = ScenarioDataManager()
        for idx in range(default_config["num_scenarios"]):
            engine.seed(idx)
            map = ScenarioMap(map_index=idx)
            heightfield = map.get_semantic_map(size, res)
            assert heightfield.shape[0] == heightfield.shape[1] == int(size * res)
            if show:
                cv2.imshow('terrain', heightfield)
                cv2.waitKey(0)
    finally:
        close_engine()


def test_map_get_elevation_map(dir="waymo", render=False, show=False):
    default_config = ScenarioEnv.default_config()
    default_config["use_render"] = render
    default_config["debug"] = False
    default_config["debug_static_world"] = False
    default_config["data_directory"] = AssetLoader.file_path(dir, return_raw_style=False)
    # default_config["data_directory"] = AssetLoader.file_path("nuscenes", return_raw_style=False)
    # default_config["data_directory"] = "/home/shady/Downloads/test_processed"
    default_config["num_scenarios"] = 3
    engine = initialize_engine(default_config)
    try:
        size, res = 1024, 1
        engine.data_manager = ScenarioDataManager()
        for idx in range(default_config["num_scenarios"]):
            engine.seed(idx)
            map = ScenarioMap(map_index=idx)
            heightfield = map.get_height_map(size, res, extension=4)
            assert heightfield.shape[0] == heightfield.shape[1] == int(size * res)
            if show:
                cv2.imshow('terrain', heightfield)
                cv2.waitKey(0)
    finally:
        close_engine()


if __name__ == "__main__":
    # test_map_get_semantic_map("waymo", render=False, show=True)
    test_map_get_elevation_map("waymo", render=False, show=True)



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from metadrive.render_pipeline.rplibs.six import iteritems
from metadrive.render_pipeline.rplibs.yaml import load_yaml_file

from direct.stdpy.file import open

from metadrive.render_pipeline.rpcore.rpobject import RPObject
from metadrive.render_pipeline.rpcore.gui.pipe_viewer import PipeViewer
from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.util.shader_input_blocks import SimpleInputBlock, GroupedInputBlock
from metadrive.render_pipeline.rpcore.stages.update_previous_pipes_stage import UpdatePreviousPipesStage


class StageManager(RPObject):
    """ This manager takes a list of RenderStages and puts them into an order,
    while connecting the different pipes, inputs, ubos and defines. """
    def __init__(self, pipeline):
        """ Constructs the stage manager """
        RPObject.__init__(self)
        self.stages = []
        self.inputs = {}
        self.pipes = {}
        self.input_blocks = []
        self.previous_pipes = {}
        self.future_bindings = []
        self.defines = {}
        self.pipeline = pipeline
        self.created = False

        self._load_stage_order()

        # Register the manager so the pipe viewer can read our data
        PipeViewer.register_stage_mgr(self)

    def _load_stage_order(self):
        """ Loads the order of all stages from the stages.yaml configuration
        file """
        orders = load_yaml_file("/$$rpconfig/stages.yaml")
        if "global_stage_order" not in orders:
            self.error("Could not load stage order, root key does not exist!")
            return
        self._stage_order = orders["global_stage_order"]

    def add_stage(self, stage):
        """ Adds a new stage """
        if stage.stage_id not in self._stage_order:
            self.error("The stage type", stage.debug_name, "is not registered yet! Please add it to the StageManager!")
            return

        if self.created:
            self.error("Cannot attach stage, stages are already created!")
            return

        self.stages.append(stage)

    def get_stage(self, stage_class):
        """ Returns a handle to an instantiated stage """
        for stage in self.stages:
            if stage.__class__.__name__ == stage_class:
                return stage

    def _prepare_stages(self):
        """ Prepares all stages by removing disabled stages and sorting stages
        by order """
        self.debug("Preparing stages ..")

        # Remove all disabled stages
        to_remove = []
        for stage in self.stages:
            if stage.disabled:
                to_remove.append(stage)

        for stage in to_remove:
            self.stages.remove(stage)

        self.stages.sort(key=lambda stage: self._stage_order.index(stage.stage_id))

    def _bind_pipes_to_stage(self, stage):
        """ Sets all required pipes on a stage """
        for pipe in stage.required_pipes:

            # Check if there is an input block named like the pipe
            if pipe in self.input_blocks:
                self.input_blocks[pipe].bind_to(stage)
                continue

            if pipe.startswith("PreviousFrame::"):
                # Special case: Pipes from the previous frame. We assume those
                # pipes have the same size as the window and a format of
                # F_rgba16. Could be subject to change.
                pipe_name = pipe.split("::")[-1]
                if pipe_name not in self.previous_pipes:
                    tex_format = "RGBA16"

                    # XXX: Assuming we have a depth texture whenever "depth"
                    # occurs in the textures name
                    if "depth" in pipe_name.lower():
                        tex_format = "R32"

                    pipe_tex = Image.create_2d("Prev-" + pipe_name, 0, 0, tex_format)
                    pipe_tex.clear_image()
                    self.previous_pipes[pipe_name] = pipe_tex
                stage.set_shader_input("Previous_" + pipe_name, self.previous_pipes[pipe_name])
                continue

            elif pipe.startswith("FuturePipe::"):
                # Special case: Future Pipes which are not available yet.
                # They will contain the unmodified data from the last
                # frame.
                pipe_name = pipe.split("::")[-1]
                self.debug("Awaiting future pipe", pipe_name)
                self.future_bindings.append((pipe_name, stage))
                continue

            if pipe not in self.pipes:
                self.fatal("Pipe '" + pipe + "' is missing for", stage)
                return False

            pipe_value = self.pipes[pipe]
            if isinstance(pipe_value, list) or isinstance(pipe_value, tuple):
                stage.set_shader_input(pipe, *pipe_value)
            else:
                stage.set_shader_input(pipe, pipe_value)
        return True

    def _bind_inputs_to_stage(self, stage):
        """ Binds all inputs including common inputs to the given stage """
        common_inputs = ["mainCam", "mainRender", "MainSceneData", "TimeOfDay"]
        for input_binding in stage.required_inputs + common_inputs:
            if input_binding not in self.inputs and \
               input_binding not in self.input_blocks:
                self.error("Input", input_binding, "is missing for", stage)
                continue

            if input_binding in self.inputs:
                stage.set_shader_input(input_binding, self.inputs[input_binding])
            elif input_binding in self.input_blocks:
                self.input_blocks[input_binding].bind_to(stage)
            else:
                assert False, "Input binding not in inputs and not in blocks!"
        return True

    def _register_stage_result(self, stage):
        """ Registers all produced pipes, inputs and defines from the given
        stage, so they can be used by later stages. """
        for pipe_name, pipe_data in (iteritems)(stage.produced_pipes):
            if isinstance(pipe_data, (SimpleInputBlock, GroupedInputBlock)):
                self.input_blocks[pipe_name] = pipe_data
                continue
            self.pipes[pipe_name] = pipe_data

        for define_name, data in iteritems(stage.produced_defines):
            if define_name in self.defines:
                self.warn("Stage", stage, "overrides define", define_name)
            self.defines[define_name] = data

        for input_name, data in iteritems(stage.produced_inputs):
            if input_name in self.inputs:
                self.warn("Stage", stage, "overrides input", input_name)

            if isinstance(data, (SimpleInputBlock, GroupedInputBlock)):
                self.input_blocks[input_name] = data
                continue

            self.inputs[input_name] = data

    def _create_previous_pipes(self):
        """ Creates a target for each last-frame's pipe, any pipe starting
        with the prefix 'Previous::' has to be stored and copied each frame. """
        if self.previous_pipes:
            self._prev_stage = UpdatePreviousPipesStage(self.pipeline)
            for prev_pipe, prev_tex in iteritems(self.previous_pipes):

                if prev_pipe not in self.pipes:
                    self.error(
                        "Attempted to use previous frame data from pipe", prev_pipe,
                        "- however, that pipe was never created!"
                    )
                    return False

                # Tell the stage to transfer the data from the current pipe to
                # the current texture
                self._prev_stage.add_transfer(self.pipes[prev_pipe], prev_tex)
            self._prev_stage.create()
            self._prev_stage.set_dimensions()
            self.stages.append(self._prev_stage)

    def _apply_future_bindings(self):
        """ Applies all future bindings. At this point all pipes and
        inputs should be present """
        for pipe, stage in self.future_bindings:
            if pipe not in self.pipes:
                self.error("Could not bind future pipe:", pipe, "not present!")
                continue
            stage.set_shader_input(pipe, self.pipes[pipe])
        self.future_bindings = []

    def setup(self):
        """ Setups the stages """
        self.debug("Setup stages ..")
        self.created = True

        # Convert input blocks so we can access them in a better way
        self.input_blocks = {block.name: block for block in self.input_blocks}
        self._prepare_stages()

        for stage in self.stages:
            stage.create()
            stage.handle_window_resize()

            # Rely on the methods to print an appropriate error message
            if not self._bind_pipes_to_stage(stage):
                continue
            if not self._bind_inputs_to_stage(stage):
                continue

            self._register_stage_result(stage)
        self._create_previous_pipes()
        self._apply_future_bindings()

    def reload_shaders(self):
        """ This pass sets the shaders to all passes and also generates the
        shader configuration """
        self.write_autoconfig()
        for stage in self.stages:
            stage.reload_shaders()

    def update(self):
        """ Calls the update method for each registered stage. Inactive stages
        are skipped. """
        for stage in self.stages:
            if stage.active:
                stage.update()

    def handle_window_resize(self):
        """ Method to get called when the window got resized. Propagates the
        resize event to all registered stages """
        for stage in self.stages:
            stage.handle_window_resize()

    def write_autoconfig(self):
        """ Writes the shader auto config, based on the defines specified by the
        different stages """
        self.debug("Writing shader config")

        # Generate autoconfig as string
        output = "#pragma once\n\n"
        output += "// Autogenerated by the render pipeline\n"
        output += "// Do not edit! Your changes will be lost.\n\n"

        for key, value in sorted(iteritems(self.defines)):
            if isinstance(value, bool):
                value = 1 if value else 0
            output += "#define " + key + " " + str(value) + "\n"

        try:
            with open("/$$rptemp/$$pipeline_shader_config.inc.glsl", "w") as handle:
                handle.write(output)
        except IOError as msg:
            self.error("Error writing shader autoconfig:", msg)



import math
from typing import Tuple

import numpy as np

from metadrive.component.lane.pg_lane import PGLane
from metadrive.constants import DrivableAreaProperty
from metadrive.constants import PGLineType
from metadrive.utils.math import wrap_to_pi, norm, Vector


class CircularLane(PGLane):
    """A lane going in circle arc."""
    def __init__(
        self,
        center: Vector,
        radius: float,
        start_phase: float,
        angle: float,
        clockwise: bool = True,
        width: float = PGLane.DEFAULT_WIDTH,
        line_types: Tuple[PGLineType, PGLineType] = (PGLineType.BROKEN, PGLineType.BROKEN),
        forbidden: bool = False,
        speed_limit: float = 1000,
        priority: int = 0
    ) -> None:
        assert angle > 0, "Angle should be greater than 0"
        super().__init__()
        self.set_speed_limit(speed_limit)
        self.center = Vector(center)
        self.radius = radius
        self._clock_wise = clockwise
        self.start_phase = start_phase
        self.end_phase = self.start_phase + (-angle if self.is_clockwise() else angle)
        self.angle = angle
        self.direction = -1 if clockwise else 1
        self.width = width
        self.line_types = line_types
        self.forbidden = forbidden
        self.priority = priority

        self.length = self.radius * (self.end_phase - self.start_phase) * self.direction
        assert self.length > 0, "end_phase should > (<) start_phase if anti-clockwise (clockwise)"
        self.start = self.position(0, 0)
        self.end = self.position(self.length, 0)

    def update_properties(self):
        self.length = self.radius * (self.end_phase - self.start_phase) * self.direction
        assert self.length > 0, "end_phase should > (<) start_phase if anti-clockwise (clockwise)"
        self.start = self.position(0, 0)
        self.end = self.position(self.length, 0)

    # def position(self, longitudinal: float, lateral: float) -> np.ndarray:
    def position(self, longitudinal: float, lateral: float) -> Vector:
        phi = self.direction * longitudinal / self.radius + self.start_phase
        # return self.center + (self.radius - lateral * self.direction) * np.array([math.cos(phi), math.sin(phi)])
        # return self.center + (self.radius - lateral * self.direction) * Vector((math.cos(phi), math.sin(phi)))
        return self.center + (self.radius + lateral * self.direction) * Vector((math.cos(phi), math.sin(phi)))

    def heading_theta_at(self, longitudinal: float) -> float:
        phi = self.direction * longitudinal / self.radius + self.start_phase
        psi = phi + math.pi / 2 * self.direction
        return psi

    def width_at(self, longitudinal: float) -> float:
        return self.width

    def local_coordinates(self, position: Tuple[float, float]) -> Tuple[float, float]:
        delta_x = position[0] - self.center[0]
        delta_y = position[1] - self.center[1]
        phi = math.atan2(delta_y, delta_x)
        phi = self.start_phase + wrap_to_pi(phi - self.start_phase)
        r = norm(delta_x, delta_y)
        longitudinal = self.direction * (phi - self.start_phase) * self.radius
        # lateral = self.direction * (self.radius - r)
        lateral = -self.direction * (self.radius - r)
        return longitudinal, lateral

    def construct_lane_in_block(self, block, lane_index):
        if self.index is None:
            self.index = lane_index
        segment_num = int(self.length / DrivableAreaProperty.LANE_SEGMENT_LENGTH)
        if segment_num == 0:
            middle = self.position(self.length / 2, 0)
            end = self.position(self.length, 0)
            theta = self.heading_theta_at(self.length / 2)
            width = self.width_at(0) + DrivableAreaProperty.SIDEWALK_LINE_DIST * 2
            self.construct_lane_segment(block, middle, width, self.length, theta, lane_index)
            return

        for i in range(segment_num):
            middle = self.position(self.length * (i + .5) / segment_num, 0)
            theta = self.heading_theta_at(self.length * (i + .5) / segment_num)
            width = self.width_at(0) + DrivableAreaProperty.SIDEWALK_LINE_DIST * 2
            length = self.length
            self._construct_lane_only_vis_segment(block, middle, width, length * 1.3 / segment_num, theta)

        self._construct_lane_only_physics_polygon(block, self.polygon)

    @property
    def polygon(self):
        if self._polygon is None:
            start_heading = self.heading_theta_at(0)
            start_dir = [math.cos(start_heading), math.sin(start_heading)]

            end_heading = self.heading_theta_at(self.length)
            end_dir = [math.cos(end_heading), math.sin(end_heading)]
            polygon = []
            longs = np.arange(0, self.length + self.POLYGON_SAMPLE_RATE, self.POLYGON_SAMPLE_RATE)
            for k, lateral in enumerate([+self.width / 2, -self.width / 2]):
                if k == 1:
                    longs = longs[::-1]
                for t, longitude in enumerate(longs):
                    point = self.position(longitude, lateral)
                    if (t == 0 and k == 0) or (t == len(longs) - 1 and k == 1):
                        # control the adding sequence
                        if k == 1:
                            # last point
                            polygon.append([point[0], point[1]])

                        # extend
                        polygon.append(
                            [
                                point[0] - start_dir[0] * self.POLYGON_SAMPLE_RATE,
                                point[1] - start_dir[1] * self.POLYGON_SAMPLE_RATE
                            ]
                        )

                        if k == 0:
                            # first point
                            polygon.append([point[0], point[1]])
                    elif (t == 0 and k == 1) or (t == len(longs) - 1 and k == 0):

                        if k == 0:
                            # second point
                            polygon.append([point[0], point[1]])

                        polygon.append(
                            [
                                point[0] + end_dir[0] * self.POLYGON_SAMPLE_RATE,
                                point[1] + end_dir[1] * self.POLYGON_SAMPLE_RATE
                            ]
                        )

                        if k == 1:
                            # third point
                            polygon.append([point[0], point[1]])
                    else:
                        polygon.append([point[0], point[1]])
            self._polygon = polygon
        return self._polygon

    def is_clockwise(self):
        return self._clock_wise



from metadrive.component.traffic_light.base_traffic_light import BaseTrafficLight
from metadrive.type import MetaDriveType


class ScenarioTrafficLight(BaseTrafficLight):
    def set_status(self, status):
        status = MetaDriveType.parse_light_status(status, simplifying=True)
        if status == MetaDriveType.LIGHT_GREEN:
            self.set_green()
        elif status == MetaDriveType.LIGHT_RED:
            self.set_red()
        elif status == MetaDriveType.LIGHT_YELLOW:
            self.set_yellow()
        elif status == MetaDriveType.LIGHT_UNKNOWN:
            self.set_unknown()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from panda3d.core import PTAInt

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.util.shader_input_blocks import SimpleInputBlock
from metadrive.render_pipeline.rpcore.pluginbase.base_plugin import BasePlugin
from metadrive.render_pipeline.rpcore.stages.cull_lights_stage import CullLightsStage

from .probe_manager import ProbeManager
from .environment_capture_stage import EnvironmentCaptureStage
from .apply_envprobes_stage import ApplyEnvprobesStage
from .cull_probes_stage import CullProbesStage


class Plugin(BasePlugin):

    name = "Environment Probes"
    author = "tobspr <tobias.springer1@gmail.com>"
    description = (
        "This plugin adds support for environment probes, containing "
        "diffuse and specular information. This enables accurate "
        "reflections, and can also be used to simulate GI."
    )
    version = "beta (!)"

    def on_stage_setup(self):
        self.probe_mgr = ProbeManager()
        self.probe_mgr.resolution = self.get_setting("probe_resolution")
        self.probe_mgr.diffuse_resolution = self.get_setting("diffuse_probe_resolution")
        self.probe_mgr.max_probes = self.get_setting("max_probes")
        self.probe_mgr.init()
        self._setup_stages()

    def _setup_stages(self):
        """ Setups all stages """
        # Create the stage to generate and update the cubemaps
        self.capture_stage = self.create_stage(EnvironmentCaptureStage)
        self.capture_stage.resolution = self.probe_mgr.resolution
        self.capture_stage.diffuse_resolution = self.probe_mgr.diffuse_resolution
        self.capture_stage.storage_tex = self.probe_mgr.cubemap_storage
        self.capture_stage.storage_tex_diffuse = self.probe_mgr.diffuse_storage

        # Create the stage to cull the cubemaps
        self.cull_stage = self.create_stage(CullProbesStage)

        # Create the stage to apply the cubemaps
        self.apply_stage = self.create_stage(ApplyEnvprobesStage)

        if self.is_plugin_enabled("scattering"):
            self.capture_stage.required_pipes += ["ScatteringIBLSpecular", "ScatteringIBLDiffuse"]

        if self.is_plugin_enabled("pssm"):
            self.capture_stage.required_pipes += ["PSSMSceneSunShadowMapPCF"]
            self.capture_stage.required_inputs += ["PSSMSceneSunShadowMVP"]

        self._setup_inputs()

    def _setup_inputs(self):
        """ Sets all required inputs """
        self.pta_probes = PTAInt.empty_array(1)

        # Construct the UBO which stores all environment probe data
        self.data_ubo = SimpleInputBlock("EnvProbes")
        self.data_ubo.add_input("num_probes", self.pta_probes)
        self.data_ubo.add_input("cubemaps", self.probe_mgr.cubemap_storage)
        self.data_ubo.add_input("diffuse_cubemaps", self.probe_mgr.diffuse_storage)
        self.data_ubo.add_input("dataset", self.probe_mgr.dataset_storage)
        self._pipeline.stage_mgr.input_blocks.append(self.data_ubo)

        # Use the UBO in light culling
        CullLightsStage.required_inputs.append("EnvProbes")

    def on_pre_render_update(self):
        if self._pipeline.task_scheduler.is_scheduled("envprobes_select_and_cull"):
            self.probe_mgr.update()
            self.pta_probes[0] = self.probe_mgr.num_probes
            probe = self.probe_mgr.find_probe_to_update()
            if probe:
                probe.last_update = Globals.clock.get_frame_count()
                self.capture_stage.active = True
                self.capture_stage.set_probe(probe)

                if self.is_plugin_enabled("pssm"):
                    self.get_plugin_instance("pssm").scene_shadow_stage.request_focus(
                        probe.bounds.get_center(), probe.bounds.get_radius()
                    )
            else:
                self.capture_stage.active = False



from metadrive.constants import TerminationState
from metadrive.envs import MetaDriveEnv
from metadrive.utils import setup_logger


def test_reward_cost_done():
    rewards = dict(
        success_reward=1111,
        out_of_road_penalty=2222,
        crash_vehicle_penalty=3333,
        crash_object_penalty=4444,
        out_of_road_cost=5555,
        crash_vehicle_cost=6666,
        crash_object_cost=7777,
    )

    # Success
    # config = rewards.copy()
    # config["map"] = "S"
    # config["traffic_density"] = 0
    # try:
    #     env = MetaDriveEnv(config=config)
    #     env.reset()
    #     for _ in range(1000):
    #         o, r, tm, tc, i = env.step([0, 1])
    #         if tm or tc:
    #             break
    #     assert i[TerminationState.SUCCESS]
    #     assert i["cost"] == 0
    #     assert r == rewards["success_reward"]
    # finally:
    #     if "env" in locals():
    #         env.close()

    # Out of road
    # config = rewards.copy()
    # config["map"] = "S"
    # config["traffic_density"] = 0
    # try:
    #     env = MetaDriveEnv(config=config)
    #     env.reset()
    #     for _ in range(1000):
    #         o, r, tm, tc, i = env.step([1, 1])
    #         if tm or tc:
    #             break
    #     assert i[TerminationState.OUT_OF_ROAD]
    #     assert i["cost"] == rewards["out_of_road_cost"]
    #     assert r == -rewards["out_of_road_penalty"]
    # finally:
    #     if "env" in locals():
    #         env.close()

    # Crash vehicle
    config = rewards.copy()
    config["map"] = "SSS"
    config["traffic_density"] = 20
    try:
        env = MetaDriveEnv(config=config)
        env.reset()
        epr = 0
        for _ in range(1000):
            o, r, tm, tc, i = env.step([0, 1])
            epr += r
            # print("R: {}, Accu R: {}".format(r, epr))
            if tm or tc:
                epr = 0
                break
        assert i[TerminationState.CRASH]
        assert i[TerminationState.CRASH_VEHICLE]
        assert i["cost"] == rewards["crash_vehicle_cost"]
        assert r == -rewards["crash_vehicle_penalty"]
    finally:
        if "env" in locals():
            env.close()

    # # Crash object
    # config = rewards.copy()
    # config["map"] = "SSS"
    # config["traffic_density"] = 0
    # config['accident_prob'] = 1.0
    # # config["use_render"] = True
    # config.update({"_debug_crash_object": True})
    # try:
    #     env = MetaDriveEnv(config=config)
    #     env.reset()
    #     for _ in range(1000):
    #         o, r, tm, tc, i = env.step([0, 1])
    #         if tm or tc:
    #             break
    #     assert i[TerminationState.CRASH]
    #     assert i[TerminationState.CRASH_OBJECT]
    #     assert i["cost"] == rewards["crash_object_cost"]
    #     assert r == -rewards["crash_object_penalty"]
    # finally:
    #     if "env" in locals():
    #         env.close()


if __name__ == '__main__':
    setup_logger(True)
    test_reward_cost_done()



from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger

if __name__ == "__main__":
    setup_logger(True)
    env = MetaDriveEnv(
        {
            "num_scenarios": 10,
            "traffic_density": 0.15,
            "traffic_mode": "hybrid",
            "start_seed": 74,
            # "_disable_detector_mask":True,
            # "debug_physics_world": True,
            # "debug": True,
            # "global_light": False,
            # "debug_static_world": True,
            "show_interface": False,
            "cull_scene": False,
            "random_spawn_lane_index": False,
            "random_lane_width": False,
            # "image_observation": True,
            # "controller": "joystick",
            # "show_coordinates": True,
            "random_agent_model": False,
            "manual_control": True,
            "use_render": True,
            "accident_prob": 1,
            "decision_repeat": 5,
            "daytime": "19:00",
            "interface_panel": [],
            "need_inverse_traffic": False,
            "rgb_clip": True,
            "map": "CX",
            # "agent_policy": ExpertPolicy,
            "random_traffic": False,
            # "random_lane_width": True,
            "driving_reward": 1.0,
            # "pstats": True,
            "force_destroy": False,
            # "show_skybox": False,
            "show_fps": False,
            "render_pipeline": True,
            # "camera_dist": 8,
            "window_size": (1600, 900),
            "camera_dist": 9,
            # "camera_pitch": 30,
            # "camera_height": 1,
            # "camera_smooth": False,
            # "camera_height": -1,
            "vehicle_config": {
                "enable_reverse": False,
                # "vehicle_model": "xl",
                # "rgb_camera": (1024, 1024),
                # "spawn_velocity": [8.728615581032535, -0.24411703918728195],
                "spawn_velocity_car_frame": True,
                # "image_source": "depth_camera",
                # "random_color": True
                # "show_lidar": True,
                "spawn_lane_index": None,
                # "destination":"2R1_3_",
                # "show_side_detector": True,
                # "show_lane_line_detector": True,
                # "side_detector": dict(num_lasers=2, distance=50),
                # "lane_line_detector": dict(num_lasers=2, distance=50),
                # "show_line_to_navi_mark": True,
                "show_navi_mark": False,
                # "show_dest_mark": True
            },
        }
    )
    import time

    speed = 8

    def acc_speed():
        global speed
        speed *= 2

    def de_speed():
        global speed
        speed /= 2

    def lower_terrain():
        pos = env.engine.terrain._mesh_terrain.getPos()
        env.engine.terrain._mesh_terrain.set_pos(pos[0], pos[1], pos[2] - speed)

    def lift_terrain():
        pos = env.engine.terrain._mesh_terrain.getPos()
        env.engine.terrain._mesh_terrain.set_pos(pos[0], pos[1], pos[2] + speed)

    init_state = {
        'position': (40.82264362985734, -509.3641208712943),
        'heading': -89.41878393159747,
        'velocity': [8.728615581032535, -0.24411703918728195],
        'valid': True
    }

    start = time.time()

    o, _ = env.reset()
    if env.config["render_pipeline"]:
        env.engine.accept("5", env.engine.render_pipeline.reload_shaders)
        env.engine.accept("7", acc_speed)
        env.engine.accept("8", de_speed)
        env.engine.accept("9", lift_terrain)
        env.engine.accept("0", lower_terrain)
    # env.main_camera.set_follow_lane(True)
    # env.vehicle.get_camera("rgb_camera").save_image(env.vehicle)
    # for line in env.engine.coordinate_line:
    #     line.reparentTo(env.vehicle.origin)
    # env.vehicle.set_velocity([5, 0], in_local_frame=True)
    print(time.time() - start)
    for s in range(1, 100000):
        # env.vehicle.set_velocity([1, 0], in_local_frame=True)
        o, r, tm, tc, info = env.step([0, 0])
        if tm:
            s = time.time()
            env.reset()
            print(time.time() - s)

        # env.vehicle.set_pitch(-np.pi/4)
        # [0.09231533, 0.491018, 0.47076905, 0.7691619, 0.5, 0.5, 1.0, 0.0, 0.48037243, 0.8904728, 0.81229943, 0.7317231, 1.0, 0.85320455, 0.9747932, 0.65675277, 0.0, 0.5, 0.5]
        # else:
        # if s % 100 == 0:
        #     env.close()
        #     env.reset()
        # info["fuel"] = env.vehicle.energy_consumption
        # env.render(
        #     text={
        #         # "heading_diff": env.vehicle.heading_diff(env.vehicle.lane),
        #         # "lane_width": env.vehicle.lane.width,
        #         # "lane_index": env.vehicle.lane_index,
        #         # "lateral": env.vehicle.lane.local_coordinates(env.vehicle.position),
        #         "current_seed": env.current_seed
        #     }
        # )
        # if tm or tc:
        #     env.reset()
        # # assert env.observation_space.contains(o)
        # if (s + 1) % 100 == 0:
        #     # print(
        #         "Finish {}/10000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
        #             s + 1,f
        #             time.time() - start, (s + 1) / (time.time() - start)
        #         )
        #     )
        # if tm or tc:
        # #     # env.close()
        # #     # print(len(env.engine._spawned_objects))
        # env.reset()
# [0.09231525, 0.4910181, 0.470769, 0.7691619, 0.5, 0.5, 1.0, 0.0, 0.4803721,

# 0.81229913, 0.8904725, 0.7317231, 1.0, 0.85320455,
# 0.6567526, 0.9747927, 0.0, 0.5, 0.5]



from metadrive.component.buildings.base_building import BaseBuilding
from metadrive.type import MetaDriveType
from metadrive.engine.asset_loader import AssetLoader
from metadrive.utils.pg.utils import generate_static_box_physics_body


class TollGateBuilding(BaseBuilding):
    BUILDING_LENGTH = 10
    BUILDING_HEIGHT = 5
    HEIGHT = BUILDING_HEIGHT
    MASS = 0

    def __init__(self, lane, position, heading_theta, random_seed):
        super(TollGateBuilding, self).__init__(position, heading_theta, lane, random_seed)
        air_wall = generate_static_box_physics_body(
            self.BUILDING_LENGTH,
            lane.width,
            self.BUILDING_HEIGHT / 2,
            object_id=self.id,
            type_name=MetaDriveType.BUILDING
        )
        self.lane_width = lane.width
        self.add_body(air_wall)

        self.set_position(position, 0)
        self.set_heading_theta(heading_theta)

        if self.render:
            building_model = self.loader.loadModel(AssetLoader.file_path("models", "tollgate", "booth.gltf"))
            gate_model = self.loader.loadModel(AssetLoader.file_path("models", "tollgate", "gate.gltf"))
            building_model.setH(90)
            building_model.reparentTo(self.origin)
            gate_model.reparentTo(self.origin)

    @property
    def top_down_length(self):
        return self.BUILDING_LENGTH

    @property
    def top_down_width(self):
        return 3

    @property
    def WIDTH(self):
        return self.lane_width

    @property
    def LENGTH(self):
        return self.BUILDING_LENGTH



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

import math

from panda3d.core import Vec3

from metadrive.render_pipeline.rpcore.pluginbase.base_plugin import BasePlugin

from .scattering_stage import ScatteringStage
from .scattering_envmap_stage import ScatteringEnvmapStage
from .godray_stage import GodrayStage


class Plugin(BasePlugin):
    name = "Atmospheric Scattering"
    author = "tobspr <tobias.springer1@gmail.com>"
    description = (
        "This plugin adds support for Atmospheric Scattering, and a "
        "single sun, based on the work from Eric Bruneton. It also "
        "adds support for atmospheric fog."
    )
    version = "1.2"

    def on_pipeline_created(self):
        self.scattering_model.load()
        self.scattering_model.compute()

    def on_stage_setup(self):
        self.display_stage = self.create_stage(ScatteringStage)
        self.envmap_stage = self.create_stage(ScatteringEnvmapStage)

        if self.get_setting("enable_godrays"):
            self.godray_stage = self.create_stage(GodrayStage)

        # Load scattering method
        method = self.get_setting("scattering_method")

        self.debug("Loading scattering method for '" + method + "'")

        if method == "eric_bruneton":
            from .scattering_methods import ScatteringMethodEricBruneton
            self.scattering_model = ScatteringMethodEricBruneton(self)
        elif method == "hosek_wilkie":
            from .scattering_methods import ScatteringMethodHosekWilkie
            self.scattering_model = ScatteringMethodHosekWilkie(self)  # noqa # pylint: disable=redefined-variable-type
        else:
            self.error("Unrecognized scattering method!")

    @property
    def sun_vector(self):
        """ Returns the sun vector """
        sun_altitude = self.get_daytime_setting("sun_altitude")
        sun_azimuth = self.get_daytime_setting("sun_azimuth")
        theta = (90 - sun_altitude) / 180.0 * math.pi
        phi = sun_azimuth / 180.0 * math.pi
        sun_vector = Vec3(math.sin(theta) * math.cos(phi), math.sin(theta) * math.sin(phi), math.cos(theta))
        return sun_vector

    def on_pre_render_update(self):
        self.envmap_stage.active = self._pipeline.task_scheduler.is_scheduled("scattering_update_envmap")

    def on_shader_reload(self):
        self.scattering_model.compute()



from metadrive import MetaDriveEnv
from metadrive.component.vehicle.vehicle_type import VaryingDynamicsVehicle, DefaultVehicle
from metadrive.engine import initialize_engine, close_engine


def test_varying_dynamics_vehicle():
    try:

        env = MetaDriveEnv()
        env.reset()
        default_config = env.config

        # close_engine()
        # initialize_engine(default_config)

        v_config = default_config.copy()["vehicle_config"]

        v_config["need_navigation"] = False

        v = VaryingDynamicsVehicle(v_config, random_seed=0)
        ref_v = DefaultVehicle(v_config, random_seed=0)
        assert v.WIDTH == ref_v.WIDTH
        assert v.LENGTH == ref_v.LENGTH
        assert v.HEIGHT == ref_v.HEIGHT

        for width in [ref_v.WIDTH, 2, 3, 4]:
            for height in [ref_v.HEIGHT, 1, 2]:
                for length in [ref_v.LENGTH, 4, 6, 9, 13, 15]:
                    for friction in [0.2, 0.8, 1.0, 1.2, 1.5, 2.0]:
                        v.reset(
                            vehicle_config={
                                "width": width,
                                "height": height,
                                "length": length,
                                "wheel_friction": friction
                            }
                        )
                        assert v.WIDTH == width
                        assert v.LENGTH == length
                        assert v.HEIGHT == height
                        for wheel in v.wheels:
                            assert abs(wheel.getFrictionSlip() - friction) < 1e-5, (wheel.getFrictionSlip(), friction)

    finally:
        close_engine()


if __name__ == "__main__":
    test_varying_dynamics_vehicle()



from metadrive.utils.opendrive.parser import parse_opendrive
from metadrive.utils.opendrive.link_index import LinkIndex
from lxml import etree


def get_lane_width(lane):
    if not len(lane.widths) <= 1:
        print("too many lane width value! move to logging warning")
    if len(lane.widths) == 1:
        width = lane.widths[0].polynomial_coefficients[0]
        # assert sum(lane.widths[0].polynomial_coefficients) == self.width, "Only support fixed lane width"
    else:
        # TODO LQY: remove 4.0
        width = lane.roadMark.get("width", 4.0)
    return float(width) if isinstance(width, str) else width


def get_lane_id(lane):
    section_id = lane.lane_section.idx
    road_id = lane.lane_section.parentRoad.id
    lane_id = lane.id
    return "{}-{}-{}".format(road_id, section_id, lane_id)


def load_opendrive_map(path):
    # Load road network and print some statistics
    try:
        fh = open(path, "r")
        openDriveXml = parse_opendrive(etree.parse(fh).getroot())
        fh.close()
        return openDriveXml
    except (etree.XMLSyntaxError) as e:
        print("XML Syntax Error: {}".format(e))
        return None
    except (TypeError, AttributeError, ValueError) as e:
        print("Value Error: {}".format(e))
        return None


if __name__ == "__main__":
    map = load_opendrive_map("C:\\Users\\x1\\Desktop\\Town01.xodr")
    link_index = LinkIndex()
    link_index.create_from_opendrive(map)
    print(link_index)



import gymnasium as gym
import numpy as np

from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.obs.observation_base import ObservationBase
from metadrive.obs.state_obs import StateObservation

_cuda_enable = True
try:
    import cupy as cp
except ImportError:
    _cuda_enable = False


class ImageStateObservation(ObservationBase):
    """
    Use ego state info, navigation info and front cam image/top down image as input
    The shape needs special handling
    """
    IMAGE = "image"
    STATE = "state"

    def __init__(self, config):
        super(ImageStateObservation, self).__init__(config)
        self.img_obs = ImageObservation(config, config["vehicle_config"]["image_source"], config["rgb_clip"])
        self.state_obs = StateObservation(config)

    @property
    def observation_space(self):
        return gym.spaces.Dict(
            {
                self.IMAGE: self.img_obs.observation_space,
                self.STATE: self.state_obs.observation_space
            }
        )

    def observe(self, vehicle: BaseVehicle):
        return {self.IMAGE: self.img_obs.observe(vehicle), self.STATE: self.state_obs.observe(vehicle)}


class ImageObservation(ObservationBase):
    """
    Use only image info as input
    """
    STACK_SIZE = 3  # use continuous 3 image as the input

    def __init__(self, config, image_source: str, clip_rgb: bool):
        self.enable_cuda = config["image_on_cuda"]
        if self.enable_cuda:
            assert _cuda_enable, "CuPy is not enabled"
        self.STACK_SIZE = config["stack_size"]
        self.image_source = image_source
        super(ImageObservation, self).__init__(config)
        self.rgb_clip = clip_rgb
        self.state = np.zeros(self.observation_space.shape, dtype=np.float32)
        if self.enable_cuda:
            self.state = cp.asarray(self.state)

    @property
    def observation_space(self):
        shape = (self.config["sensors"][self.image_source][2], self.config["sensors"][self.image_source][1]
                 ) + ((self.STACK_SIZE, ) if self.config["rgb_to_grayscale"] else (3, self.STACK_SIZE))
        if self.rgb_clip:
            return gym.spaces.Box(-0.0, 1.0, shape=shape, dtype=np.float32)
        else:
            return gym.spaces.Box(0, 255, shape=shape, dtype=np.uint8)

    def observe(self, vehicle):
        new_obs = self.engine.get_sensor(self.image_source).get_pixels_array(vehicle, self.rgb_clip)
        self.state = cp.roll(self.state, -1, axis=-1) if self.enable_cuda else np.roll(self.state, -1, axis=-1)
        self.state[..., -1] = new_obs
        return self.state

    def get_image(self):
        return self.state.copy()[:, :, -1]

    def reset(self, env, vehicle=None):
        """
        Clear stack
        :param env: MetaDrive
        :param vehicle: BaseVehicle
        :return: None
        """
        self.state = np.zeros(self.observation_space.shape, dtype=np.float32)
        if self.enable_cuda:
            self.state = cp.asarray(self.state)



import copy

import geopandas as gpd
import numpy as np
from nuscenes import NuScenes
from nuscenes.can_bus.can_bus_api import NuScenesCanBus
from nuscenes.eval.common.utils import quaternion_yaw
from nuscenes.map_expansion.arcline_path_utils import discretize_lane
from nuscenes.map_expansion.map_api import NuScenesMap
from pyquaternion import Quaternion
from shapely.ops import unary_union, cascaded_union
import logging
from metadrive.scenario import ScenarioDescription as SD
from metadrive.type import MetaDriveType
from metadrive.utils.nuscenes.detection_type import ALL_TYPE, HUMAN_TYPE, BICYCLE_TYPE, VEHICLE_TYPE
logger = logging.getLogger(__name__)
EGO = "ego"

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)


def get_metadrive_type(obj_type):
    meta_type = obj_type
    md_type = None
    if ALL_TYPE[obj_type] == "barrier":
        md_type = MetaDriveType.TRAFFIC_BARRIER
    elif ALL_TYPE[obj_type] == "trafficcone":
        md_type = MetaDriveType.TRAFFIC_CONE
    elif obj_type in VEHICLE_TYPE:
        md_type = MetaDriveType.VEHICLE
    elif obj_type in HUMAN_TYPE:
        md_type = MetaDriveType.PEDESTRIAN
    elif obj_type in BICYCLE_TYPE:
        md_type = MetaDriveType.CYCLIST

    # assert meta_type != MetaDriveType.UNSET and meta_type != "noise"
    return md_type, meta_type


def parse_frame(frame, nuscenes: NuScenes):
    ret = {}
    for obj_id in frame["anns"]:
        obj = nuscenes.get("sample_annotation", obj_id)
        # velocity = nuscenes.box_velocity(obj_id)[:2]
        # if np.nan in velocity:
        velocity = np.array([0.0, 0.0])
        ret[obj["instance_token"]] = {
            "position": obj["translation"],
            "obj_id": obj["instance_token"],
            "heading": quaternion_yaw(Quaternion(*obj["rotation"])),
            "rotation": obj["rotation"],
            "velocity": velocity,
            "size": obj["size"],
            "visible": obj["visibility_token"],
            "attribute": [nuscenes.get("attribute", i)["name"] for i in obj["attribute_tokens"]],
            "type": obj["category_name"]
        }
    ego_token = nuscenes.get("sample_data", frame["data"]["LIDAR_TOP"])["ego_pose_token"]
    ego_state = nuscenes.get("ego_pose", ego_token)
    ret[EGO] = {
        "position": ego_state["translation"],
        "obj_id": EGO,
        "heading": quaternion_yaw(Quaternion(*ego_state["rotation"])),
        "rotation": ego_state["rotation"],
        "type": "vehicle.car",
        "velocity": np.array([0.0, 0.0]),
        # size https://en.wikipedia.org/wiki/Renault_Zoe
        "size": [4.08, 1.73, 1.56],
    }
    return ret


def interpolate_heading(heading_data, old_valid, new_valid, num_to_interpolate=5):
    new_heading_theta = np.zeros_like(new_valid)
    for k, valid in enumerate(old_valid[:-1]):
        if abs(valid) > 1e-1 and abs(old_valid[k + 1]) > 1e-1:
            diff = (heading_data[k + 1] - heading_data[k] + np.pi) % (2 * np.pi) - np.pi
            # step = diff
            interpolate_heading = np.linspace(heading_data[k], heading_data[k] + diff, 6)
            new_heading_theta[k * num_to_interpolate:(k + 1) * num_to_interpolate] = interpolate_heading[:-1]
        elif abs(valid) > 1e-1 and abs(old_valid[k + 1]) < 1e-1:
            new_heading_theta[k * num_to_interpolate:(k + 1) * num_to_interpolate] = heading_data[k]
    new_heading_theta[-1] = heading_data[-1]
    return new_heading_theta * new_valid


def _interpolate_one_dim(data, old_valid, new_valid, num_to_interpolate=5):
    new_data = np.zeros_like(new_valid)
    for k, valid in enumerate(old_valid[:-1]):
        if abs(valid) > 1e-1 and abs(old_valid[k + 1]) > 1e-1:
            diff = data[k + 1] - data[k]
            # step = diff
            interpolate_data = np.linspace(data[k], data[k] + diff, num_to_interpolate + 1)
            new_data[k * num_to_interpolate:(k + 1) * num_to_interpolate] = interpolate_data[:-1]
        elif abs(valid) > 1e-1 and abs(old_valid[k + 1]) < 1e-1:
            new_data[k * num_to_interpolate:(k + 1) * num_to_interpolate] = data[k]
    new_data[-1] = data[-1]
    return new_data * new_valid


def interpolate(origin_y, valid, new_valid):
    if len(origin_y.shape) == 1:
        ret = _interpolate_one_dim(origin_y, valid, new_valid)
    elif len(origin_y.shape) == 2:
        ret = []
        for dim in range(origin_y.shape[-1]):
            new_y = _interpolate_one_dim(origin_y[..., dim], valid, new_valid)
            new_y = np.expand_dims(new_y, axis=-1)
            ret.append(new_y)
        ret = np.concatenate(ret, axis=-1)
    else:
        raise ValueError("Y has shape {}, Can not interpolate".format(origin_y.shape))
    return ret


def get_tracks_from_frames(nuscenes: NuScenes, scene_info, frames, num_to_interpolate=5):
    episode_len = len(frames)
    # Fill tracks
    all_objs = set()
    for frame in frames:
        all_objs.update(frame.keys())
    tracks = {
        k: dict(
            type=MetaDriveType.UNSET,
            state=dict(
                position=np.zeros(shape=(episode_len, 3)),
                heading=np.zeros(shape=(episode_len, )),
                velocity=np.zeros(shape=(episode_len, 2)),
                valid=np.zeros(shape=(episode_len, )),
                length=np.zeros(shape=(episode_len, 1)),
                width=np.zeros(shape=(episode_len, 1)),
                height=np.zeros(shape=(episode_len, 1))
            ),
            metadata=dict(track_length=episode_len, type=MetaDriveType.UNSET, object_id=k, original_id=k)
        )
        for k in list(all_objs)
    }

    tracks_to_remove = set()

    for frame_idx in range(episode_len):
        # Record all agents' states (position, velocity, ...)
        for id, state in frames[frame_idx].items():
            # Fill type
            md_type, meta_type = get_metadrive_type(state["type"])
            tracks[id]["type"] = md_type
            tracks[id][SD.METADATA]["type"] = meta_type
            if md_type is None or md_type == MetaDriveType.UNSET:
                tracks_to_remove.add(id)
                continue

            tracks[id]["type"] = md_type
            tracks[id][SD.METADATA]["type"] = meta_type

            # Introducing the state item
            tracks[id]["state"]["position"][frame_idx] = state["position"]
            tracks[id]["state"]["heading"][frame_idx] = state["heading"]
            tracks[id]["state"]["velocity"][frame_idx] = tracks[id]["state"]["velocity"][frame_idx]
            tracks[id]["state"]["valid"][frame_idx] = 1

            tracks[id]["state"]["length"][frame_idx] = state["size"][1]
            tracks[id]["state"]["width"][frame_idx] = state["size"][0]
            tracks[id]["state"]["height"][frame_idx] = state["size"][2]

            tracks[id]["metadata"]["original_id"] = id
            tracks[id]["metadata"]["object_id"] = id

    for track in tracks_to_remove:
        track_data = tracks.pop(track)
        obj_type = track_data[SD.METADATA]["type"]
        print("\nWARNING: Can not map type: {} to any MetaDrive Type".format(obj_type))

    new_episode_len = (episode_len - 1) * num_to_interpolate + 1

    # interpolate
    interpolate_tracks = {}
    for id, track, in tracks.items():
        interpolate_tracks[id] = copy.deepcopy(track)
        interpolate_tracks[id]["metadata"]["track_length"] = new_episode_len

        # valid first
        new_valid = np.zeros(shape=(new_episode_len, ))
        if track["state"]["valid"][0]:
            new_valid[0] = 1
        for k, valid in enumerate(track["state"]["valid"][1:], start=1):
            if valid:
                if abs(new_valid[(k - 1) * num_to_interpolate] - 1) < 1e-2:
                    start_idx = (k - 1) * num_to_interpolate + 1
                else:
                    start_idx = k * num_to_interpolate
                new_valid[start_idx:k * num_to_interpolate + 1] = 1
        interpolate_tracks[id]["state"]["valid"] = new_valid

        # position
        interpolate_tracks[id]["state"]["position"] = interpolate(
            track["state"]["position"], track["state"]["valid"], new_valid
        )
        if id == "ego":
            # We can get it from canbus
            canbus = NuScenesCanBus(dataroot=nuscenes.dataroot)
            imu_pos = np.asarray([state["pos"] for state in canbus.get_messages(scene_info["name"], "pose")[::5]])
            interpolate_tracks[id]["state"]["position"][:len(imu_pos)] = imu_pos

        # velocity
        interpolate_tracks[id]["state"]["velocity"] = interpolate(
            track["state"]["velocity"], track["state"]["valid"], new_valid
        )
        vel = interpolate_tracks[id]["state"]["position"][1:] - interpolate_tracks[id]["state"]["position"][:-1]
        interpolate_tracks[id]["state"]["velocity"][:-1] = vel[..., :2] / 0.1
        for k, valid in enumerate(new_valid[1:], start=1):
            if valid == 0 or not valid or abs(valid) < 1e-2:
                interpolate_tracks[id]["state"]["velocity"][k] = np.array([0., 0.])
                interpolate_tracks[id]["state"]["velocity"][k - 1] = np.array([0., 0.])
        # speed outlier check
        max_vel = np.max(np.linalg.norm(interpolate_tracks[id]["state"]["velocity"], axis=-1))
        assert max_vel < 50, "Abnormal velocity!"
        if max_vel > 30:
            print("\nWARNING: Too large peed for {}: {}".format(id, max_vel))

        # heading
        # then update position
        new_heading = interpolate_heading(track["state"]["heading"], track["state"]["valid"], new_valid)
        interpolate_tracks[id]["state"]["heading"] = new_heading
        if id == "ego":
            # We can get it from canbus
            canbus = NuScenesCanBus(dataroot=nuscenes.dataroot)
            imu_heading = np.asarray(
                [
                    quaternion_yaw(Quaternion(state["orientation"]))
                    for state in canbus.get_messages(scene_info["name"], "pose")[::5]
                ]
            )
            interpolate_tracks[id]["state"]["heading"][:len(imu_heading)] = imu_heading

        for k, v in track["state"].items():
            if k in ["valid", "heading", "position", "velocity"]:
                continue
            else:
                interpolate_tracks[id]["state"][k] = interpolate(v, track["state"]["valid"], new_valid)
        # if id == "ego":
        # ego is valid all time, so we can calculate the velocity in this way

    return interpolate_tracks


def get_map_features(scene_info, nuscenes: NuScenes, map_center, radius=250, points_distance=1):
    """
    Extract map features from nuscenes data. The objects in specified region will be returned. Sampling rate determines
    the distance between 2 points when extracting lane center line.
    """
    ret = {}
    map_name = nuscenes.get("log", scene_info["log_token"])["location"]
    map_api = NuScenesMap(dataroot=nuscenes.dataroot, map_name=map_name)

    layer_names = [
        # "line",
        # "polygon",
        # "node",
        'drivable_area',
        'road_segment',
        'road_block',
        'lane',
        # 'ped_crossing',
        # 'walkway',
        # 'stop_line',
        # 'carpark_area',
        'lane_connector',
        'road_divider',
        'lane_divider',
        'traffic_light'
    ]
    map_objs = map_api.get_records_in_radius(map_center[0], map_center[1], radius, layer_names)

    # build map boundary
    polygons = []
    # for id in map_objs["drivable_area"]:
    #     seg_info = map_api.get("drivable_area", id)
    #     assert seg_info["token"] == id
    #     for polygon_token in seg_info["polygon_tokens"]:
    #         polygon = map_api.extract_polygon(polygon_token)
    #         polygons.append(polygon)
    for id in map_objs["road_segment"]:
        seg_info = map_api.get("road_segment", id)
        assert seg_info["token"] == id
        polygon = map_api.extract_polygon(seg_info["polygon_token"])
        polygons.append(polygon)
    for id in map_objs["road_block"]:
        seg_info = map_api.get("road_block", id)
        assert seg_info["token"] == id
        polygon = map_api.extract_polygon(seg_info["polygon_token"])
        polygons.append(polygon)
    polygons = [geom if geom.is_valid else geom.buffer(0) for geom in polygons]
    # logger.warning("Stop using boundaries! Use exterior instead!")
    boundaries = gpd.GeoSeries(unary_union(polygons)).boundary.explode(index_parts=True)
    for idx, boundary in enumerate(boundaries[0]):
        block_points = np.array(list(i for i in zip(boundary.coords.xy[0], boundary.coords.xy[1])))
        id = "boundary_{}".format(idx)
        ret[id] = {SD.TYPE: MetaDriveType.LINE_SOLID_SINGLE_WHITE, SD.POLYLINE: block_points}

    for id in map_objs["lane_divider"]:
        line_info = map_api.get("lane_divider", id)
        assert line_info["token"] == id
        line = map_api.extract_line(line_info["line_token"]).coords.xy
        line = [[line[0][i], line[1][i]] for i in range(len(line[0]))]
        ret[id] = {SD.TYPE: MetaDriveType.LINE_BROKEN_SINGLE_WHITE, SD.POLYLINE: line}

    for id in map_objs["road_divider"]:
        line_info = map_api.get("road_divider", id)
        assert line_info["token"] == id
        line = map_api.extract_line(line_info["line_token"]).coords.xy
        line = [[line[0][i], line[1][i]] for i in range(len(line[0]))]
        ret[id] = {SD.TYPE: MetaDriveType.LINE_SOLID_SINGLE_YELLOW, SD.POLYLINE: line}

    for id in map_objs["lane"]:
        lane_info = map_api.get("lane", id)
        assert lane_info["token"] == id
        boundary = map_api.extract_polygon(lane_info["polygon_token"]).boundary.xy
        boundary_polygon = [[boundary[0][i], boundary[1][i]] for i in range(len(boundary[0]))]
        # boundary_polygon += [[boundary[0][i], boundary[1][i]] for i in range(len(boundary[0]))]
        ret[id] = {
            SD.TYPE: MetaDriveType.LANE_SURFACE_STREET,
            SD.POLYLINE: discretize_lane(map_api.arcline_path_3[id], resolution_meters=points_distance),
            SD.POLYGON: boundary_polygon,
            # TODO Add speed limit if needed
            "speed_limit_kmh": 100
        }

    for id in map_objs["lane_connector"]:
        lane_info = map_api.get("lane_connector", id)
        assert lane_info["token"] == id
        # boundary = map_api.extract_polygon(lane_info["polygon_token"]).boundary.xy
        # boundary_polygon = [[boundary[0][i], boundary[1][i], 0.1] for i in range(len(boundary[0]))]
        # boundary_polygon += [[boundary[0][i], boundary[1][i], 0.] for i in range(len(boundary[0]))]
        ret[id] = {
            SD.TYPE: MetaDriveType.LANE_SURFACE_STREET,
            SD.POLYLINE: discretize_lane(map_api.arcline_path_3[id], resolution_meters=points_distance),
            # SD.POLYGON: boundary_polygon,
            "speed_limit_kmh": 100
        }

    return ret


def convert_one_scenario(scene_token: str, nuscenes: NuScenes):
    """
    Data will be interpolated to 0.1s time interval, while the time interval of original key frames are 0.5s.
    """
    scenario_log_interval = 0.1
    scene_info = nuscenes.get("scene", scene_token)
    frames = []
    current_frame = nuscenes.get("sample", scene_info["first_sample_token"])
    while current_frame["token"] != scene_info["last_sample_token"]:
        frames.append(parse_frame(current_frame, nuscenes))
        current_frame = nuscenes.get("sample", current_frame["next"])
    frames.append(parse_frame(current_frame, nuscenes))
    assert current_frame["next"] == ""
    assert len(frames) == scene_info["nbr_samples"], "Number of sample mismatches! "

    result = SD()
    result[SD.ID] = scene_info["name"]
    result[SD.VERSION] = "nuscenes" + nuscenes.version
    result[SD.LENGTH] = (len(frames) - 1) * 5 + 1
    result[SD.METADATA] = {}
    result[SD.METADATA]["dataset"] = "nuscenes"
    result[SD.METADATA][SD.METADRIVE_PROCESSED] = False
    result[SD.METADATA]["map"] = nuscenes.get("log", scene_info["log_token"])["location"]
    result[SD.METADATA]["date"] = nuscenes.get("log", scene_info["log_token"])["date_captured"]
    result[SD.METADATA]["coordinate"] = "right-handed"
    result[SD.METADATA]["scenario_token"] = scene_token
    result[SD.METADATA]["scenario_id"] = scene_info["name"]
    result[SD.METADATA]["sample_rate"] = scenario_log_interval
    result[SD.METADATA][SD.TIMESTEP] = np.arange(0., (len(frames) - 1) * 0.5 + 0.1, 0.1)
    # interpolating to 0.1s interval
    result[SD.TRACKS] = get_tracks_from_frames(nuscenes, scene_info, frames, num_to_interpolate=5)
    result[SD.METADATA][SD.SDC_ID] = "ego"

    # TODO Traffic Light
    result[SD.DYNAMIC_MAP_STATES] = {}

    # map
    map_center = result[SD.TRACKS]["ego"]["state"]["position"][0]
    result[SD.MAP_FEATURES] = get_map_features(scene_info, nuscenes, map_center, 250)

    return result



from metadrive.envs.metadrive_env import MetaDriveEnv
from metadrive.utils import setup_logger


def test_traffic_respawn(vis=False):
    setup_logger(vis)

    env = MetaDriveEnv({"num_scenarios": 1, "manual_control": vis, "use_render": vis, "traffic_mode": "respawn"})
    env.reset()
    try:
        for i in range(1, 3000):
            env.step([0, 0])
            current_v = set(env.engine.traffic_manager.vehicles)
            for v in list(env.engine.traffic_manager.traffic_vehicles) + [env.vehicle]:
                if v is env.vehicle:
                    current_v.discard(v)
                else:
                    current_v.discard(v)
            assert len(current_v) == 0, "vehicles didn't release"
            assert len(env.engine.traffic_manager.vehicles) - len(env.engine.traffic_manager.traffic_vehicles) == 1, \
                "vehicles didn't release"
    finally:
        env.close()


if __name__ == '__main__':
    test_traffic_respawn(vis=False)



import logging
from typing import Tuple

import math
import numpy as np

from metadrive.component.lane.point_lane import PointLane
from metadrive.utils.opendrive.elements.geometry import Line, Arc
from metadrive.utils.opendrive.map_load import get_lane_id


class OpenDriveLane(PointLane):
    ARC_SEGMENT_LENGTH = 1  # m
    """An OpenDrive Lane"""
    def __init__(self, width, lane_data) -> None:
        self.lane_data = lane_data
        self.width = width
        geos = self.lane_data.parentRoad.planView._geometries
        self._initialize_geometry(geos)
        self.single_side = lane_data.lane_section.singleSide

        self._section_index = lane_data.lane_section.idx
        self.index = get_lane_id(lane_data)
        self.roadMark_color = lane_data.roadMark.get("color", None)
        self.roadMark_type = lane_data.roadMark.get("type", None)
        self.roadMark_material = lane_data.roadMark.get("material", None)

    def _initialize_geometry(self, geos):
        points = []
        for geo in geos:
            if isinstance(geo, Line):
                heading = geo.heading
                length = geo.length
                start = geo.start_position
                points.append(start)
                # if geo is geos[-1]:
                #     # last geo
                end = start + np.array([np.cos(heading) * length, np.sin(heading) * length])
                points.append(end)
            elif isinstance(geo, Arc):
                continue
                arc_points = self._arc_interpolate(1 / geo.curvature, geo.length, geo.start_position, geo.heading)
                points += arc_points
                if geo is not geos[-1] and len(arc_points) > 1:
                    # not last geo
                    points.pop()
            else:
                raise ValueError("Only support Line and Arc, currently")
        super(OpenDriveLane, self).__init__(points, self.width)

    def _arc_interpolate(self, radius, length, start_position, start_phase):
        ret = []
        arc_degree = ((radius / length) + np.pi) % np.pi * 2 - np.pi
        start_degree = start_phase
        origin = start_position - np.array([np.cos(start_degree) * radius, np.sin(start_degree) * radius])
        num_to_seg = math.floor(length / self.ARC_SEGMENT_LENGTH)
        if num_to_seg == 0:
            degree = start_degree - arc_degree
            degree = (degree + np.pi) % np.pi * 2 - np.pi
            ret.append(np.array([np.cos(degree) * radius, np.sin(degree) * radius]) + origin)
            ret.append(degree)
        # else:
        #     each_seg_degree = arc_degree / num_to_seg
        #     for i in range(num_to_seg + 1):
        #         degree = i * each_seg_degree + start_degree
        #         degree = (degree + np.pi) % np.pi * 2- np.pi
        #         ret.append(np.array([np.cos(degree) * radius, np.sin(degree) * radius]) + origin)
        return ret

    def is_lane_line(self):
        return True if self.roadMark_type in [
            "solid", "broken", "broken broken", "solid solid", "solid broken", "broken solid"
        ] else False

    def destroy(self):
        self.width = None
        # lane line will be processed separately
        self.start = None
        self.end = None
        super(OpenDriveLane, self).destroy()



#!/usr/bin/env python3
# Copyright (c) 2020 kitsune.ONE team.

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.

# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.

# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import argparse


def parse_args():
    parser = argparse.ArgumentParser()

    parser.add_argument('inputs', type=str, help='Input .blend file path.', nargs='*')
    parser.add_argument(
        '-b', '--background', action='store_true', required=False, help="Blender's argument placeholder."
    )
    parser.add_argument('-P', '--python', type=str, required=False, help="Blender's argument placeholder.")
    parser.add_argument('-o', '--output', type=str, required=False, help='Output .egg file path.')
    parser.add_argument(
        '-e', '--export', type=str, required=False, default='scene', help='Export type: scene/animation/collision/all'
    )
    parser.add_argument('-r', '--render', type=str, required=False, default='default', help='Render type: default/rp.')
    parser.add_argument('-x', '--exec', type=str, required=False, help='Internal script name to execute.')
    parser.add_argument('-a', '--action', type=str, required=False, help='Action name to export.')
    parser.add_argument('-sp', '--speed', type=float, required=False, help='Animations speed scale.')
    parser.add_argument('-sc', '--scale', type=float, required=False, help='Geom scale.')
    parser.add_argument('-m', '--merge', action='store_true', help='Merge objects and meshes inside the collection.')
    parser.add_argument(
        '-k', '--keep', action='store_true', help='Keep the original objects and meshes before merging.'
    )
    parser.add_argument('-nuv', '--no-extra-uv', action='store_true', help="Don't export extra non-primary UV.")
    parser.add_argument(
        '-nmat', '--no-materials', action='store_true', help="Don't export materials (skips textures aswell)."
    )
    parser.add_argument('-ntex', '--no-textures', action='store_true', help="Don't export textures.")
    parser.add_argument(
        '-etex', '--empty-textures', action='store_true', help="Use placeholder images for empty texture slots."
    )
    parser.add_argument(
        '-sorg', '--set-origin', action='store_true', help="Set origin to center of bounds for collisions."
    )

    return parser.parse_args()


def main():
    args = parse_args()
    from panda3d.egg import EggData
    from metadrive.third_party.kitsunetsuki.exporter.egg import EggExporter
    e = EggExporter(args)
    out = e.convert()

    if args.output:
        if isinstance(out, EggData):
            out.write_egg(args.output)
        else:
            out.write(args.output)
    else:
        print(out)


if __name__ == '__main__':
    main()



from metadrive.component.lane.straight_lane import StraightLane
from metadrive.component.pgblock.create_pg_block_utils import ExtendStraightLane, CreateRoadFrom, CreateAdverseRoad
from metadrive.component.pgblock.pg_block import PGBlock, PGBlockSocket
from metadrive.component.road_network import Road
from metadrive.constants import PGLineType
from metadrive.component.pg_space import ParameterSpace, Parameter, BlockParameterSpace


class Straight(PGBlock):
    """
    Straight Road
    ----------------------------------------
    ----------------------------------------
    ----------------------------------------
    """
    ID = "S"
    SOCKET_NUM = 1
    PARAMETER_SPACE = ParameterSpace(BlockParameterSpace.STRAIGHT)

    def _try_plug_into_previous_block(self) -> bool:
        self.set_part_idx(0)  # only one part in simple block like straight, and curve
        para = self.get_config()
        length = para[Parameter.length]
        basic_lane = self.positive_basic_lane
        assert isinstance(basic_lane, StraightLane), "Straight road can only connect straight type"
        new_lane = ExtendStraightLane(basic_lane, length, [PGLineType.BROKEN, PGLineType.SIDE])
        start = self.pre_block_socket.positive_road.end_node
        end = self.add_road_node()
        socket = Road(start, end)
        _socket = -socket

        # create positive road
        no_cross = CreateRoadFrom(
            new_lane,
            self.positive_lane_num,
            socket,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        )
        # create negative road
        no_cross = CreateAdverseRoad(
            socket,
            self.block_network,
            self._global_network,
            ignore_intersection_checking=self.ignore_intersection_checking
        ) and no_cross
        self.add_sockets(PGBlockSocket(socket, _socket))
        return no_cross



from metadrive.envs.metadrive_env import MetaDriveEnv

if __name__ == "__main__":
    env = MetaDriveEnv(
        {
            "map": 30,
            "num_scenarios": 1,
            "traffic_density": 0.1,
            "pstats": True,
            "traffic_mode": "respawn"
        }
    )

    o, _ = env.reset()
    for i in range(1, 10000):
        # print(i)
        o, r, tm, tc, info = env.step([0, 0])
    env.close()



import copy

from metadrive.component.map.pg_map import PGMap
from metadrive.component.pgblock.first_block import FirstPGBlock
from metadrive.component.pgblock.roundabout import Roundabout
from metadrive.component.road_network import Road
from metadrive.envs.marl_envs.multi_agent_metadrive import MultiAgentMetaDrive
from metadrive.manager.pg_map_manager import PGMapManager
from metadrive.manager.spawn_manager import SpawnManager
from metadrive.utils import Config

MARoundaboutConfig = dict(
    spawn_roads=[
        Road(FirstPGBlock.NODE_2, FirstPGBlock.NODE_3),
        -Road(Roundabout.node(1, 0, 2), Roundabout.node(1, 0, 3)),
        -Road(Roundabout.node(1, 1, 2), Roundabout.node(1, 1, 3)),
        -Road(Roundabout.node(1, 2, 2), Roundabout.node(1, 2, 3)),
    ],
    map_config=dict(exit_length=60, lane_num=2),
    top_down_camera_initial_x=95,
    top_down_camera_initial_y=15,
    top_down_camera_initial_z=120,
    num_agents=40,
)


class MARoundaboutMap(PGMap):
    def _generate(self):
        length = self.config["exit_length"]

        parent_node_path, physics_world = self.engine.worldNP, self.engine.physics_world
        assert len(self.road_network.graph) == 0, "These Map is not empty, please create a new map to read config"

        # Build a first-block
        last_block = FirstPGBlock(
            self.road_network,
            self.config[self.LANE_WIDTH],
            self.config[self.LANE_NUM],
            parent_node_path,
            physics_world,
            length=length
        )
        self.blocks.append(last_block)

        # Build roundabout
        Roundabout.EXIT_PART_LENGTH = length
        last_block = Roundabout(
            1, last_block.get_socket(index=0), self.road_network, random_seed=1, ignore_intersection_checking=False
        )
        last_block.construct_block(
            parent_node_path,
            physics_world,
            extra_config={
                "exit_radius": 10,
                "inner_radius": 30,
                "angle": 70,
                # Note: lane_num is set in config.map_config.lane_num
            }
        )
        self.blocks.append(last_block)


class MARoundaboutPGMapManager(PGMapManager):
    def reset(self):
        config = self.engine.global_config
        if len(self.spawned_objects) == 0:
            _map = self.spawn_object(MARoundaboutMap, map_config=config["map_config"], random_seed=None)
        else:
            assert len(self.spawned_objects) == 1, "It is supposed to contain one map in this manager"
            _map = self.spawned_objects.values()[0]
        self.load_map(_map)
        self.current_map.spawn_roads = config["spawn_roads"]


# Note: This class allow one agent accessing its neighbors' observation.
# But we never use this functionality in formal experiment. Therefore we just deprecate it for
# clarity!
# class LidarStateObservationMARound(ObservationBase):
#     def __init__(self, vehicle_config):
#         self.state_obs = StateObservation(vehicle_config)
#         super(LidarStateObservationMARound, self).__init__(vehicle_config)
#         self.state_length = list(self.state_obs.observation_space.shape)[0]
#         self.cloud_points = None
#         self.detected_objects = None
#
#     @property
#     def observation_space(self):
#         shape = list(self.state_obs.observation_space.shape)
#         if self.config["lidar"]["num_lasers"] > 0 and self.config["lidar"]["distance"] > 0:
#             # Number of lidar rays and distance should be positive!
#             shape[0] += self.config["lidar"]["num_lasers"] + self.config["lidar"]["num_others"] * self.state_length
#         return gym.spaces.Box(-0.0, 1.0, shape=tuple(shape), dtype=np.float32)
#
#     def observe(self, vehicle):
#         num_others = self.config["lidar"]["num_others"]
#         state = self.state_observe(vehicle)
#         other_v_info = []
#         if vehicle.lidar.available:
#             cloud_points, detected_objects = vehicle.lidar.perceive(vehicle)
#             if self.config["lidar"]["num_others"] > 0:
#                 surrounding_vehicles = list(vehicle.lidar.get_surrounding_vehicles(detected_objects))
#                 surrounding_vehicles.sort(
#                     key=lambda v: norm(vehicle.position[0] - v.position[0], vehicle.position[1] - v.position[1])
#                 )
#                 surrounding_vehicles += [None] * num_others
#                 for tmp_v in surrounding_vehicles[:num_others]:
#                     if tmp_v is not None:
#                         tmp_v = tmp_v
#                         other_v_info += self.state_observe(tmp_v).tolist()
#                     else:
#                         other_v_info += [0] * self.state_length
#             other_v_info += self._add_noise_to_cloud_points(
#                 cloud_points,
#                 gaussian_noise=self.config["lidar"]["gaussian_noise"],
#                 dropout_prob=self.config["lidar"]["dropout_prob"]
#             )
#             self.cloud_points = cloud_points
#             self.detected_objects = detected_objects
#         self.current_observation = np.concatenate((state, np.asarray(other_v_info)))
#         return self.current_observation.astype(np.float32)
#
#     def state_observe(self, vehicle):
#         return self.state_obs.observe(vehicle)
#
#     def _add_noise_to_cloud_points(self, points, gaussian_noise, dropout_prob):
#         if gaussian_noise > 0.0:
#             points = np.asarray(points)
#             points = np.clip(points + np.random.normal(loc=0.0, scale=gaussian_noise, size=points.shape), 0.0, 1.0)
#
#         if dropout_prob > 0.0:
#             assert dropout_prob <= 1.0
#             points = np.asarray(points)
#             points[np.random.uniform(0, 1, size=points.shape) < dropout_prob] = 0.0
#
#         return list(points)


class RoundaboutSpawnManager(SpawnManager):
    def update_destination_for(self, vehicle_id, vehicle_config):
        end_roads = copy.deepcopy(self.engine.global_config["spawn_roads"])
        end_road = -self.np_random.choice(end_roads)  # Use negative road!
        vehicle_config["destination"] = end_road.end_node
        return vehicle_config


class MultiAgentRoundaboutEnv(MultiAgentMetaDrive):
    @staticmethod
    def default_config() -> Config:
        return MultiAgentMetaDrive.default_config().update(MARoundaboutConfig, allow_add_new_key=True)

    def setup_engine(self):
        super(MultiAgentRoundaboutEnv, self).setup_engine()
        self.engine.update_manager("spawn_manager", RoundaboutSpawnManager())
        self.engine.update_manager("map_manager", MARoundaboutPGMapManager())


def _draw():
    env = MultiAgentRoundaboutEnv()
    o, _ = env.reset()
    from metadrive.utils.draw_top_down_map import draw_top_down_map
    import matplotlib.pyplot as plt

    plt.imshow(draw_top_down_map(env.current_map))
    plt.show()
    env.close()


def _expert():
    env = MultiAgentRoundaboutEnv(
        {
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 240,
                    "num_others": 4,
                    "distance": 50
                },
            },
            "use_AI_protector": True,
            "save_level": 1.,
            "debug_physics_world": True,

            # "use_render": True,
            "debug": True,
            "manual_control": True,
            "num_agents": 4,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step(env.action_space.sample())
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        tm.update({"total_r": total_r, "episode length": ep_s})
        # env.render(text=d)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis_debug_respawn():
    env = MultiAgentRoundaboutEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            "debug_physics_world": True,
            "use_render": True,
            "debug": False,
            "manual_control": True,
            "num_agents": 40,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        action = {k: [0.0, .0] for k in env.vehicles.keys()}
        o, r, tm, tc, info = env.step(action)
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # d.update({"total_r": total_r, "episode length": ep_s})
        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height
        }
        env.render(text=render_text)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _vis():
    env = MultiAgentRoundaboutEnv(
        {
            "horizon": 100000,
            "vehicle_config": {
                "lidar": {
                    "num_lasers": 72,
                    "num_others": 0,
                    "distance": 40
                },
                "show_lidar": False,
            },
            "use_render": True,
            "debug": False,
            "manual_control": True,
            "num_agents": -1,
        }
    )
    o, _ = env.reset()
    total_r = 0
    ep_s = 0
    for i in range(1, 100000):
        o, r, tm, tc, info = env.step({k: [0, .0] for k in env.vehicles.keys()})
        for r_ in r.values():
            total_r += r_
        ep_s += 1
        # tm.update({"total_r": total_r, "episode length": ep_s})
        render_text = {
            "total_r": total_r,
            "episode length": ep_s,
            "cam_x": env.main_camera.camera_x,
            "cam_y": env.main_camera.camera_y,
            "cam_z": env.main_camera.top_down_camera_height
        }
        env.render(text=render_text)
        if tm["__all__"]:
            print(
                "Finish! Current step {}. Group Reward: {}. Average reward: {}".format(
                    i, total_r, total_r / env.agent_manager.next_agent_count
                )
            )
            # break
        if len(env.vehicles) == 0:
            total_r = 0
            print("Reset")
            env.reset()
    env.close()


def _profile():
    import time
    env = MultiAgentRoundaboutEnv({"num_agents": 40})
    obs, _ = env.reset()
    start = time.time()
    for s in range(10000):
        o, r, tm, tc, i = env.step(env.action_space.sample())

        # mask_ratio = env.engine.detector_mask.get_mask_ratio()
        # print("Mask ratio: ", mask_ratio)

        if all(tm.values()):
            env.reset()
        if (s + 1) % 100 == 0:
            print(
                "Finish {}/10000 simulation steps. Time elapse: {:.4f}. Average FPS: {:.4f}".format(
                    s + 1,
                    time.time() - start, (s + 1) / (time.time() - start)
                )
            )
    print(f"(MetaDriveEnv) Total Time Elapse: {time.time() - start}")


def _long_run():
    # Please refer to test_ma_roundabout_reward_done_alignment()
    _out_of_road_penalty = 3
    env = MultiAgentRoundaboutEnv(
        {
            "num_agents": 40,
            "vehicle_config": {
                "lidar": {
                    "num_others": 8
                },
            },
            **dict(
                out_of_road_penalty=_out_of_road_penalty,
                crash_vehicle_penalty=1.333,
                crash_object_penalty=11,
                crash_vehicle_cost=13,
                crash_object_cost=17,
                out_of_road_cost=19,
            )
        }
    )
    try:
        obs, _ = env.reset()
        assert env.observation_space.contains(obs)
        for step in range(10000):
            act = env.action_space.sample()
            o, r, tm, tc, i = env.step(act)
            if step == 0:
                assert not any(tm.values())

            if any(tm.values()):
                print("Current Done: {}\nReward: {}".format(tm, r))
                for kkk, ddd in tm.items():
                    if ddd and kkk != "__all__":
                        print("Info {}: {}\n".format(kkk, i[kkk]))
                print("\n")

            for kkk, rrr in r.items():
                if rrr == -_out_of_road_penalty:
                    assert tm[kkk]

            if (step + 1) % 200 == 0:
                print(
                    "{}/{} Agents: {} {}\nO: {}\nR: {}\nD: {}\nI: {}\n\n".format(
                        step + 1, 10000, len(env.vehicles), list(env.vehicles.keys()),
                        {k: (oo.shape, oo.mean(), oo.min(), oo.max())
                         for k, oo in o.items()}, r, tm, i
                    )
                )
            if tm["__all__"]:
                print('Current step: ', step)
                break
    finally:
        env.close()


if __name__ == "__main__":
    # _draw()
    _vis()
    # _vis_debug_respawn()
    # _profile()
    # _long_run()
    # pygame_replay("round", MultiAgentRoundaboutEnv)



import os

import open3d as o3d
import matplotlib.pyplot as plt
import numpy as np

if __name__ == '__main__':
    vis = o3d.visualization.Visualizer()
    vis.create_window(width=1600, height=900, visible=True)

    # 8bit 1channel or 8bit 3channel, and depth image is: 16bit 1channel image
    pcds = []
    # for h in range(-180, 180, 20):
    h = 0
    dir = "C:\\Users\\x1\\Desktop\\neurips_2023\\sensor_video\\depth"
    files = os.listdir(dir)
    if "output.mp4" in files:
        files.remove("output.mp4")
    s = sorted(files, key=lambda x: int(x[6:-4]))
    for k, file in enumerate(s):
        raw_depth = o3d.io.read_image(os.path.join(dir, file))
        raw_depth = np.array(raw_depth)[..., 0]
        raw_depth *= 255
        raw_depth = raw_depth.astype(np.uint16)
        raw_depth = o3d.geometry.Image(raw_depth)
        # extrinsic = np.array([[np.cos(h), -np.sin(h), 0, 0],
        #                       [np.sin(h), np.cos(h), 0, 0],
        #                       [0, 0, 1, 0],
        #                       [0, 0, 0, 1]])
        extrinsic = np.array([[1, 0, 0, 0], [0, np.cos(h), -np.sin(h), 0], [0, np.sin(h), np.cos(h), 0], [0, 0, 0, 1]])
        intrinsic = o3d.camera.PinholeCameraIntrinsic(o3d.camera.PinholeCameraIntrinsicParameters.PrimeSenseDefault)
        intrinsic.set_intrinsics(1600, 900, 1.866, 2.066, 800, 450)
        pcd = o3d.geometry.PointCloud.create_from_depth_image(
            depth=raw_depth,
            intrinsic=intrinsic,
            extrinsic=extrinsic,
            project_valid_depth_only=True,
            stride=5,
            # depth_scale=10000.0,
            # depth_trunc=10000
        )
        # pcd.transform([[1, 0, 0, 0], [0, -1, 0, 0], [0, 0, -1, 0], [0, 0, 0, 1]])
        r = pcd.get_rotation_matrix_from_xyz((0, 0, np.pi))
        pcd = pcd.rotate(r, center=(0, 0, 0))
        # pcd = pcd.scale(10, center=(0, 0, 0))
        # pcd = pcd.scale(10, center=(0, 0, 0))
        # pcd.translate(np.array([0,0,-100]))
        vis.add_geometry(pcd)
        ctr = vis.get_view_control()
        assert id(ctr) == id(vis.get_view_control())  # assertion error.
        # ctr.set_up([0, -1, 0])
        # ctr.set_front([0, 0, -1])
        # ctr.set_lookat([1, 0, 0])
        ctr.set_zoom(0.25)
        #
        # camera_parameters = ctr.convert_to_pinhole_camera_parameters()
        # ex = np.array(camera_parameters.extrinsic)
        # ex[2][3]+=10
        # camera_parameters.extrinsic = ex
        # ctr.convert_from_pinhole_camera_parameters(camera_parameters)

        # vis.update_geometry(pcd)
        vis.poll_events()
        vis.update_renderer()

        vis.capture_screen_image("img_{}.png".format(k))
        vis.remove_geometry(pcd)

    vis.destroy_window()

    # CX_DEPTH=0
    # CY_DEPTH=0
    # FX_DEPTH=300
    # FY_DEPTH=300
    # raw_depth = o3d.io.read_image('debug_{}.jpg'.format(0))
    # raw_depth = np.array(raw_depth)[..., 0]
    # raw_depth = raw_depth.astype(np.uint8)
    # width = 800
    # height = 600
    # # o3d.visualization.draw_geometries(pcds)
    #
    # jj = np.tile(range(width), height)
    # ii = np.repeat(range(height), width)
    # # Compute constants:
    # xx = (jj - CX_DEPTH) / FX_DEPTH
    # yy = (ii - CY_DEPTH) / FY_DEPTH
    # # transform depth image to vector of z:
    # length = height * width
    # z = raw_depth.reshape(height * width)
    # # compute point cloud
    # pcd = np.dstack((xx * z, yy * z, z)).reshape((length, 3))
    # pcd_o3d = o3d.geometry.PointCloud()  # create point cloud object
    # pcd_o3d.points = o3d.utility.Vector3dVector(pcd)  # set pcd_np as the point cloud points
    # # Visualize:
    # o3d.visualization.draw_geometries([pcd_o3d])



"""

Main PyYAML importer script. Provides functions to load YAML files as dictionaries.

"""

from __future__ import print_function

import sys
import collections
from direct.stdpy.file import open
from metadrive.render_pipeline.rpcore.rpobject import RPObject

# Import different PyYaml versions depending on the used python version
if sys.version_info < (3, 0):
    from .yaml_py2 import load as yaml_load
    from .yaml_py2 import YAMLError, SafeLoader
else:
    from .yaml_py3 import load as yaml_load
    from .yaml_py3 import YAMLError, SafeLoader

__all__ = ["load_yaml_file", "load_yaml_file_flat"]


def load_yaml_file(filename):
    """ This method is a wrapper arround yaml_load, and provides error checking """

    import time
    start = time.process_time()

    try:
        with open(filename, "r") as handle:
            parsed_yaml = yaml_load(handle, Loader=SafeLoader)
    except IOError as msg:
        RPObject.global_error("YAMLLoader", "Could not find or open file:", filename)
        RPObject.global_error("YAMLLoader", msg)
        raise Exception("Failed to load YAML file: File not found")
    except YAMLError as msg:
        RPObject.global_error("YAMLLoader", "Invalid yaml-syntax in file:", filename)
        RPObject.global_error("YAMLLoader", msg)
        raise Exception("Failed to load YAML file: Invalid syntax")

    duration = (time.process_time() - start) * 1000.0

    # Optionally print out profiling information
    # print("Took", round(duration, 2), "ms to load", filename)

    return parsed_yaml


def __flatten(d, parent_key=''):
    """ Internal method to flatten a dictionary """
    items = []
    for k, v in d.items():
        try:
            items.extend(__flatten(v, '{}{}.'.format(parent_key, k)).items())
        except AttributeError:
            items.append(('{}{}'.format(parent_key, k), v))
    return dict(items)


def load_yaml_file_flat(filename):
    """ Behaves like load_yaml_file, but instead of creating nested dictionaries
    it connects keys via '.' """
    return __flatten(load_yaml_file(filename))



import copy
import logging
from collections import defaultdict, namedtuple

import math
import numpy as np

from metadrive.component.vehicle.base_vehicle import BaseVehicle
from metadrive.envs import MetaDriveEnv

detect_result = namedtuple("detect_result", "cloud_points detected_objects")


class DetectorMask:
    def __init__(self, num_lasers: int, max_span: float, max_distance: float = 1e6):
        logging.warning(
            "This class is deprecated, the new lidar is implemented with this optimization, and"
            "can be easily extended to 3d"
        )
        self.num_lasers = num_lasers
        self.angle_delta = 360 / self.num_lasers
        # self.max_span = max_span
        self.half_max_span_square = (max_span / 2)**2
        self.masks = defaultdict(lambda: np.zeros((self.num_lasers, ), dtype=bool))
        # self.max_distance = max_distance + max_span
        self.max_distance_square = (max_distance + max_span)**2

    def update_mask(self, position_dict: dict, heading_dict: dict, is_target_vehicle_dict: dict):
        assert set(position_dict.keys()) == set(heading_dict.keys()) == set(is_target_vehicle_dict.keys())
        if not position_dict:
            return

        for k in self.masks.keys():
            self.masks[k].fill(False)

        for k, is_target in is_target_vehicle_dict.items():
            if is_target:
                self.masks[k]  # Touch to create entry

        keys = list(position_dict.keys())
        for c1, k1 in enumerate(keys[:-1]):
            for c2, k2 in enumerate(keys[c1 + 1:]):

                if (not is_target_vehicle_dict[k1]) and (not is_target_vehicle_dict[k2]):
                    continue

                pos1 = position_dict[k1]
                pos2 = position_dict[k2]
                head1 = heading_dict[k1]
                head2 = heading_dict[k2]

                diff = (pos2[0] - pos1[0], pos2[1] - pos1[1])
                dist_square = diff[0]**2 + diff[1]**2
                if dist_square < self.half_max_span_square:
                    if is_target_vehicle_dict[k1]:
                        self._mark_all(k1)
                    if is_target_vehicle_dict[k2]:
                        self._mark_all(k2)
                    continue

                if dist_square > self.max_distance_square:
                    continue

                span = None
                if is_target_vehicle_dict[k1]:
                    span = math.asin(math.sqrt(self.half_max_span_square / dist_square))
                    # relative heading of v2's center when compared to v1's center
                    relative_head = math.atan2(diff[1], diff[0])
                    head_in_1 = relative_head - head1
                    head_in_1_max = head_in_1 + span
                    head_in_1_min = head_in_1 - span
                    head_1_max = np.rad2deg(head_in_1_max)
                    head_1_min = np.rad2deg(head_in_1_min)
                    self._mark_this_range(head_1_min, head_1_max, name=k1)

                if is_target_vehicle_dict[k2]:
                    if span is None:
                        span = math.asin(math.sqrt(self.half_max_span_square / dist_square))
                    diff2 = (-diff[0], -diff[1])
                    # relative heading of v2's center when compared to v1's center
                    relative_head2 = math.atan2(diff2[1], diff2[0])
                    head_in_2 = relative_head2 - head2
                    head_in_2_max = head_in_2 + span
                    head_in_2_min = head_in_2 - span
                    head_2_max = np.rad2deg(head_in_2_max)
                    head_2_min = np.rad2deg(head_in_2_min)
                    self._mark_this_range(head_2_min, head_2_max, name=k2)

    def _mark_this_range(self, small_angle, large_angle, name):
        # We use clockwise to determine small and large angle.
        # For example, if you wish to fill 355 deg to 5 deg, then small_angle is 355, large_angle is 5.
        small_angle = small_angle % 360
        large_angle = large_angle % 360

        assert 0 <= small_angle <= 360
        assert 0 <= large_angle <= 360

        small_index = math.floor(small_angle / self.angle_delta)
        large_index = math.ceil(large_angle / self.angle_delta)
        if large_angle < small_angle:  # We are in the scenario like small=355, large=5
            self.masks[name][small_index:] = True
            self.masks[name][:large_index + 1] = True
        else:
            self.masks[name][small_index:large_index + 1] = True

    def _mark_all(self, name):
        self.masks[name].fill(True)

    def get_mask(self, name):
        assert name in self.masks, "It seems that you have not initialized the mask for vehicle {} yet!".format(name)
        return self.masks[name]

    def clear(self):
        self.masks.clear()

    def get_mask_ratio(self):
        total = 0
        masked = 0
        for k, v in self.masks.items():
            total += v.size
            masked += v.sum()
        return masked / total


def _line_intersect(theta, center, point1, point2, maximum: float = 10000):
    """
    theta: the direction of the laser
    center: the center of the source of laser
    point1: one point of the line intersection
    point2: another poing of the line intersection
    maximum: the return value if no intersection
    """
    x0, y0 = center
    x1, y1 = point1
    x2, y2 = point2
    dx1 = math.cos(theta)
    dy1 = math.sin(theta)
    dx2 = x2 - x1
    dy2 = y2 - y1
    DET = -dx1 * dy2 + dy1 * dx2
    if abs(DET) < 1e-9:
        return maximum
    r = 1.0 / DET * (-dy2 * (x1 - x0) + dx2 * (y1 - y0))
    s = 1.0 / DET * (-dy1 * (x1 - x0) + dx1 * (y1 - y0))
    if 0 - 1e-5 < s < 1 + 1e-5 and r >= 0:
        return r
    return maximum


def _search_angle(point1, point2, num_lasers, start, heading, perceive_distance: float = 50):
    laser_heading = np.arange(0, num_lasers) * 2 * np.pi / num_lasers + heading
    result = []
    for laser_index in range(num_lasers):
        ret = _line_intersect(theta=laser_heading[laser_index], center=start, point1=point1, point2=point2, maximum=1e8)
        assert 0 <= ret < 1e9
        if ret <= 1e8:
            result.append(ret / 1e8)
        else:
            result.append(1.0)
    return np.asarray(result)


def _test_mask(mask, stick_1_heading_deg, stick_2_heading_deg, max_span, stick1_x, stick2_x):
    stick_1_heading = np.deg2rad(stick_1_heading_deg)
    stick_2_heading = np.deg2rad(stick_2_heading_deg)
    stick_1_heading = stick_1_heading % (2 * np.pi)
    stick_2_heading = stick_2_heading % (2 * np.pi)

    stick1_pos = (stick1_x, 0)
    stick2_pos = (stick2_x, 0)
    mask.update_mask(
        position_dict={
            "stick1": stick1_pos,
            "stick2": stick2_pos
        },
        heading_dict={
            "stick1": stick_1_heading,
            "stick2": stick_2_heading
        },
        is_target_vehicle_dict={
            "stick1": True,
            "stick2": True
        }
    )
    mask_1 = mask.get_mask("stick1")
    mask_2 = mask.get_mask("stick2")

    left_of_stick2 = (stick2_pos[0], -max_span / 2)
    right_of_stick2 = (stick2_pos[0], max_span / 2)

    real_mask_1 = _search_angle(
        point1=left_of_stick2,
        point2=right_of_stick2,
        num_lasers=360,
        start=stick1_pos,
        heading=stick_1_heading,
        perceive_distance=100 * max_span
    ) < 1.0
    res = np.stack([real_mask_1, mask_1])
    assert all(mask_1[real_mask_1])  # mask 1 should at least include all True of real mask.
    if abs(stick1_x - stick2_x) > max_span:
        assert sum(abs(mask_1.astype(int) - real_mask_1.astype(int))) <= 3

    left_of_stick1 = (stick1_pos[0], -max_span / 2)
    right_of_stick1 = (stick1_pos[0], max_span / 2)

    real_mask_2 = _search_angle(
        point1=left_of_stick1,
        point2=right_of_stick1,
        num_lasers=360,
        start=stick2_pos,
        heading=stick_2_heading,
        perceive_distance=100 * max_span
    ) < 1.0
    res2 = np.stack([real_mask_2, mask_2])
    assert all(mask_2[real_mask_2])  # mask 1 should at least include all True of real mask.
    if abs(stick1_x - stick2_x) > max_span:
        assert sum(abs(mask_2.astype(int) - real_mask_2.astype(int))) <= 3


def test_detector_mask():
    # A infinite long (1e7) stick 2 in front of (0.01m) stick 1.

    pos_xy = [0, -1, 1, -100, 100]
    angles = [0, 0.01, 30, 89, 90, 91, 130, 180, 181, 270, 360, 400]
    angles += [-a for a in angles]

    mask = DetectorMask(num_lasers=360, max_span=1e7)
    _test_mask(mask, -270, 30, 1e7, 0, -1)

    mask = DetectorMask(num_lasers=360, max_span=1)
    _test_mask(mask, -180, -300, 1, 0, -1)
    _test_mask(mask, -361, 270, 1, 0, -100)

    for max_span in [1e7, 1, 0.1]:
        mask = DetectorMask(num_lasers=360, max_span=max_span)
        for pos1_x in pos_xy:
            for pos2_x in pos_xy:
                angles1 = np.random.choice(angles, 5)
                angles2 = np.random.choice(angles, 5)
                for stick_1_heading_deg in angles1:
                    for stick_2_heading_deg in angles2:
                        _test_mask(mask, stick_1_heading_deg, stick_2_heading_deg, max_span, pos1_x, pos2_x)
                # print("Finish. ", max_span, pos1_x, pos2_x)


def test_detector_mask_in_lidar():
    env = MetaDriveEnv(
        {
            "traffic_density": 1.0,
            "map": "SSSSS",
            "random_traffic": False,
            "use_render": False,
            "debug": True
        }
    )
    try:
        env.reset()
        span = 2 * max(env.vehicle.WIDTH, env.vehicle.LENGTH)
        detector_mask = DetectorMask(
            env.vehicle.config.lidar.num_lasers, span, max_distance=env.vehicle.config.lidar.distance
        )
        ep_count = 0
        for tt in range(3000):
            o, r, tm, tc, i = env.step([0, 1])

            # print("We have: {} vehicles!".format(env.engine.traffic_manager.get_vehicle_num()))

            v = env.vehicle
            c_p, objs = v.lidar.perceive(v, detector_mask=None)
            old_objs = v.lidar.get_surrounding_vehicles(objs)
            old_cloud_points = np.array(copy.deepcopy(c_p))

            position_dict = {}
            heading_dict = {}
            is_target_vehicle_dict = {}
            for v in env.engine.traffic_manager.vehicles:
                position_dict[v.name] = v.position
                heading_dict[v.name] = v.heading_theta
                is_target_vehicle_dict[v.name] = True if isinstance(v, BaseVehicle) else False

            detector_mask.update_mask(
                position_dict=position_dict, heading_dict=heading_dict, is_target_vehicle_dict=is_target_vehicle_dict
            )

            real_mask = old_cloud_points != 1.0
            mask = detector_mask.get_mask(env.vehicle.name)
            stack = np.stack([old_cloud_points, real_mask, mask])
            if not all(mask[real_mask]):
                print('stop')
            assert all(mask[real_mask])  # mask 1 should at least include all True of real mask.

            print(
                "Num of true in our mask: {}, in old mask: {}. Overlap: {}. We have {} more.".format(
                    sum(mask.astype(int)), sum(real_mask.astype(int)), sum(mask[real_mask].astype(int)),
                    sum(mask.astype(int)) - sum(real_mask.astype(int))
                )
            )

            # assert sum(abs(mask.astype(int) - real_mask.astype(int))) <= 3
            v = env.vehicle
            c_p, objs = v.lidar.perceive(v)
            new_cloud_points = np.array(copy.deepcopy(c_p))
            np.testing.assert_almost_equal(old_cloud_points, new_cloud_points)

            if tm or tc:
                env.reset()
                ep_count += 1
                if ep_count == 3:
                    break
    finally:
        env.close()


if __name__ == '__main__':
    # test_detector_mask()
    test_detector_mask_in_lidar()



import logging
import os
import tempfile
from dataclasses import dataclass
from os.path import join

import numpy as np
from nuplan.common.actor_state.agent import Agent
from nuplan.common.actor_state.static_object import StaticObject
from nuplan.common.actor_state.tracked_objects_types import TrackedObjectType
from nuplan.common.maps.maps_datatypes import TrafficLightStatusType
from shapely.geometry.linestring import LineString
from shapely.geometry.multilinestring import MultiLineString

from metadrive.scenario import ScenarioDescription as SD
from metadrive.type import MetaDriveType
from metadrive.utils.coordinates_shift import nuplan_to_metadrive_vector
from metadrive.utils.math import compute_angular_velocity

logging.basicConfig(level=logging.INFO)
logger = logging.getLogger(__name__)
from metadrive.utils import is_win

try:
    import geopandas as gpd
    from nuplan.common.actor_state.state_representation import Point2D
    from nuplan.common.maps.maps_datatypes import SemanticMapLayer, StopLineType
    from shapely.ops import unary_union
    from nuplan.planning.scenario_builder.nuplan_db.nuplan_scenario import NuPlanScenario
    import hydra
    from nuplan.planning.scenario_builder.nuplan_db.nuplan_scenario import NuPlanScenario
    from nuplan.planning.script.builders.scenario_building_builder import build_scenario_builder
    from nuplan.planning.script.builders.scenario_filter_builder import build_scenario_filter
    from nuplan.planning.script.utils import set_up_common_builder
    import nuplan
except ImportError:
    logger.warning("Can not import nuplan-devkit")

logger.warning(
    "This converters will de deprecated. "
    "Use tools in ScenarioNet instead: https://github.com/metadriverse/ScenarioNet."
)

NUPLAN_PACKAGE_PATH = os.path.dirname(nuplan.__file__)
EGO = "ego"


def get_nuplan_scenarios(dataset_parameters, nuplan_package_path=NUPLAN_PACKAGE_PATH):
    """
    Return a list of nuplan scenarios according to dataset_parameters
    """
    base_config_path = os.path.join(nuplan_package_path, "planning", "script")
    simulation_hydra_paths = construct_simulation_hydra_paths(base_config_path)

    # Initialize configuration management system
    hydra.core.global_hydra.GlobalHydra.instance().clear()  # reinitialize hydra if already initialized
    hydra.initialize_config_dir(config_dir=simulation_hydra_paths.config_path)

    save_dir = tempfile.mkdtemp()
    ego_controller = 'perfect_tracking_controller'  # [log_play_back_controller, perfect_tracking_controller]
    observation = 'box_observation'  # [box_observation, idm_agents_observation, lidar_pc_observation]

    # Compose the configuration
    overrides = [
        f'group={save_dir}',
        'worker=sequential',
        f'ego_controller={ego_controller}',
        f'observation={observation}',
        f'hydra.searchpath=[{simulation_hydra_paths.common_dir}, {simulation_hydra_paths.experiment_dir}]',
        'output_dir=${group}/${experiment}',
        'metric_dir=${group}/${experiment}',
        *dataset_parameters,
    ]
    overrides.extend(
        [
            f'job_name=planner_tutorial', 'experiment=${experiment_name}/${job_name}',
            f'experiment_name=planner_tutorial'
        ]
    )

    # get config
    cfg = hydra.compose(config_name=simulation_hydra_paths.config_name, overrides=overrides)

    profiler_name = 'building_simulation'
    common_builder = set_up_common_builder(cfg=cfg, profiler_name=profiler_name)

    # Build scenario builder
    scenario_builder = build_scenario_builder(cfg=cfg)
    scenario_filter = build_scenario_filter(cfg.scenario_filter)

    # get scenarios from database
    return scenario_builder.get_scenarios(scenario_filter, common_builder.worker)


def construct_simulation_hydra_paths(base_config_path: str):
    """
    Specifies relative paths to simulation configs to pass to hydra to declutter tutorial.
    :param base_config_path: Base config path.
    :return: Hydra config path.
    """
    common_dir = "file://" + join(base_config_path, 'config', 'common')
    config_name = 'default_simulation'
    config_path = join(base_config_path, 'config', 'simulation')
    experiment_dir = "file://" + join(base_config_path, 'experiments')
    return HydraConfigPaths(common_dir, config_name, config_path, experiment_dir)


@dataclass
class HydraConfigPaths:
    """
    Stores relative hydra paths to declutter tutorial.
    """

    common_dir: str
    config_name: str
    config_path: str
    experiment_dir: str


def extract_centerline(map_obj, nuplan_center):
    path = map_obj.baseline_path.discrete_path
    points = np.array([nuplan_to_metadrive_vector([pose.x, pose.y], nuplan_center) for pose in path])
    return points


def get_points_from_boundary(boundary, center):
    path = boundary.discrete_path
    points = [(pose.x, pose.y) for pose in path]
    points = nuplan_to_metadrive_vector(points, center)
    return points


def get_line_type(nuplan_type):
    return MetaDriveType.LINE_BROKEN_SINGLE_WHITE
    # Always return broken line type
    if nuplan_type == 2:
        return MetaDriveType.LINE_SOLID_SINGLE_WHITE
    elif nuplan_type == 0:
        return MetaDriveType.LINE_BROKEN_SINGLE_WHITE
    elif nuplan_type == 3:
        return MetaDriveType.LINE_UNKNOWN
    else:
        raise ValueError("Unknown line tyep: {}".format(nuplan_type))


def extract_map_features(map_api, center, radius=250):
    ret = {}
    np.seterr(all='ignore')
    # Center is Important !
    layer_names = [
        SemanticMapLayer.LANE_CONNECTOR,
        SemanticMapLayer.LANE,
        SemanticMapLayer.CROSSWALK,
        SemanticMapLayer.INTERSECTION,
        SemanticMapLayer.STOP_LINE,
        SemanticMapLayer.WALKWAYS,
        SemanticMapLayer.CARPARK_AREA,
        SemanticMapLayer.ROADBLOCK,
        SemanticMapLayer.ROADBLOCK_CONNECTOR,

        # unsupported yet
        # SemanticMapLayer.STOP_SIGN,
        # SemanticMapLayer.DRIVABLE_AREA,
    ]
    center_for_query = Point2D(*center)
    nearest_vector_map = map_api.get_proximal_map_objects(center_for_query, radius, layer_names)
    boundaries = map_api._get_vector_map_layer(SemanticMapLayer.BOUNDARIES)
    # Filter out stop polygons in turn stop
    if SemanticMapLayer.STOP_LINE in nearest_vector_map:
        stop_polygons = nearest_vector_map[SemanticMapLayer.STOP_LINE]
        nearest_vector_map[SemanticMapLayer.STOP_LINE] = [
            stop_polygon for stop_polygon in stop_polygons if stop_polygon.stop_line_type != StopLineType.TURN_STOP
        ]
    block_polygons = []
    for layer in [SemanticMapLayer.ROADBLOCK, SemanticMapLayer.ROADBLOCK_CONNECTOR]:
        for block in nearest_vector_map[layer]:
            for lane_meta_data in block.interior_edges:
                if not hasattr(lane_meta_data, "baseline_path"):
                    continue
                if isinstance(lane_meta_data.polygon.boundary, MultiLineString):
                    # logger.warning("Stop using boundaries! Use exterior instead!")
                    boundary = gpd.GeoSeries(lane_meta_data.polygon.boundary).explode(index_parts=True)
                    sizes = []
                    for idx, polygon in enumerate(boundary[0]):
                        sizes.append(len(polygon.xy[1]))
                    points = boundary[0][np.argmax(sizes)].xy
                elif isinstance(lane_meta_data.polygon.boundary, LineString):
                    points = lane_meta_data.polygon.boundary.xy
                polygon = [[points[0][i], points[1][i]] for i in range(len(points[0]))]
                polygon = nuplan_to_metadrive_vector(polygon, nuplan_center=[center[0], center[1]])
                ret[lane_meta_data.id] = {
                    SD.TYPE: MetaDriveType.LANE_SURFACE_STREET,
                    SD.POLYLINE: extract_centerline(lane_meta_data, center),
                    SD.POLYGON: polygon
                }
                if layer == SemanticMapLayer.ROADBLOCK_CONNECTOR:
                    continue
                left = lane_meta_data.left_boundary
                if left.id not in ret:
                    line_type = get_line_type(int(boundaries.loc[[str(left.id)]]["boundary_type_fid"]))
                    if line_type != MetaDriveType.LINE_UNKNOWN:
                        ret[left.id] = {SD.TYPE: line_type, SD.POLYLINE: get_points_from_boundary(left, center)}

            if layer == SemanticMapLayer.ROADBLOCK:
                block_polygons.append(block.polygon)

    interpolygons = [block.polygon for block in nearest_vector_map[SemanticMapLayer.INTERSECTION]]
    # logger.warning("Stop using boundaries! Use exterior instead!")
    boundaries = gpd.GeoSeries(unary_union(interpolygons + block_polygons)).boundary.explode(index_parts=True)
    # boundaries.plot()
    # plt.show()
    for idx, boundary in enumerate(boundaries[0]):
        block_points = np.array(list(i for i in zip(boundary.coords.xy[0], boundary.coords.xy[1])))
        block_points = nuplan_to_metadrive_vector(block_points, center)
        id = "boundary_{}".format(idx)
        ret[id] = {SD.TYPE: MetaDriveType.LINE_SOLID_SINGLE_WHITE, SD.POLYLINE: block_points}
    np.seterr(all='warn')
    return ret


def set_light_status(status):
    if status == TrafficLightStatusType.GREEN:
        return MetaDriveType.LIGHT_GREEN
    elif status == TrafficLightStatusType.RED:
        return MetaDriveType.LIGHT_RED
    elif status == TrafficLightStatusType.YELLOW:
        return MetaDriveType.LIGHT_YELLOW
    elif status == TrafficLightStatusType.UNKNOWN:
        return MetaDriveType.LIGHT_UNKNOWN


def set_light_position(scenario, lane_id, center, target_position=8):
    lane = scenario.map_api.get_map_object(str(lane_id), SemanticMapLayer.LANE_CONNECTOR)
    assert lane is not None, "Can not find lane: {}".format(lane_id)
    path = lane.baseline_path.discrete_path
    acc_length = 0
    point = [path[0].x, path[0].y]
    for k, point in enumerate(path[1:], start=1):
        previous_p = path[k - 1]
        acc_length += np.linalg.norm([point.x - previous_p.x, point.y - previous_p.y])
        if acc_length > target_position:
            break
    return [point.x - center[0], point.y - center[1]]


def extract_traffic_light(scenario, center):
    length = scenario.get_number_of_iterations()

    frames = [
        {str(t.lane_connector_id): t.status
         for t in scenario.get_traffic_light_status_at_iteration(i)} for i in range(length)
    ]
    all_lights = set()
    for frame in frames:
        all_lights.update(frame.keys())

    lights = {
        k: {
            "type": MetaDriveType.TRAFFIC_LIGHT,
            "state": {
                SD.TRAFFIC_LIGHT_STATUS: [MetaDriveType.LIGHT_UNKNOWN] * length
            },
            SD.TRAFFIC_LIGHT_POSITION: None,
            SD.TRAFFIC_LIGHT_LANE: str(k),
            "metadata": dict(track_length=length, type=None, object_id=str(k), lane_id=str(k), dataset="nuplan")
        }
        for k in list(all_lights)
    }

    for k, frame in enumerate(frames):
        for lane_id, status in frame.items():
            lane_id = str(lane_id)
            lights[lane_id]["state"][SD.TRAFFIC_LIGHT_STATUS][k] = set_light_status(status)
            if lights[lane_id][SD.TRAFFIC_LIGHT_POSITION] is None:
                assert isinstance(lane_id, str), "Lane ID should be str"
                lights[lane_id][SD.TRAFFIC_LIGHT_POSITION] = set_light_position(scenario, lane_id, center)
                lights[lane_id][SD.METADATA][SD.TYPE] = MetaDriveType.TRAFFIC_LIGHT

    return lights


def get_traffic_obj_type(nuplan_type):
    if nuplan_type == TrackedObjectType.VEHICLE:
        return MetaDriveType.VEHICLE
    elif nuplan_type == TrackedObjectType.TRAFFIC_CONE:
        return MetaDriveType.TRAFFIC_CONE
    elif nuplan_type == TrackedObjectType.PEDESTRIAN:
        return MetaDriveType.PEDESTRIAN
    elif nuplan_type == TrackedObjectType.BICYCLE:
        return MetaDriveType.CYCLIST
    elif nuplan_type == TrackedObjectType.BARRIER:
        return MetaDriveType.TRAFFIC_BARRIER
    elif nuplan_type == TrackedObjectType.EGO:
        raise ValueError("Ego should not be in detected resukts")
    else:
        return None


def parse_object_state(obj_state, nuplan_center):
    ret = {}
    ret["position"] = nuplan_to_metadrive_vector([obj_state.center.x, obj_state.center.y], nuplan_center)
    ret["heading"] = obj_state.center.heading
    ret["velocity"] = nuplan_to_metadrive_vector([obj_state.velocity.x, obj_state.velocity.y])
    ret["valid"] = 1
    ret["length"] = obj_state.box.length
    ret["width"] = obj_state.box.width
    ret["height"] = obj_state.box.height
    return ret


def parse_ego_vehicle_state(state, nuplan_center):
    center = nuplan_center
    ret = {}
    ret["position"] = nuplan_to_metadrive_vector([state.waypoint.x, state.waypoint.y], center)
    ret["heading"] = state.waypoint.heading
    ret["velocity"] = nuplan_to_metadrive_vector([state.agent.velocity.x, state.agent.velocity.y])
    ret["angular_velocity"] = state.dynamic_car_state.angular_velocity
    ret["valid"] = 1
    ret["length"] = state.agent.box.length
    ret["width"] = state.agent.box.width
    ret["height"] = state.agent.box.height
    return ret


def parse_ego_vehicle_state_trajectory(scenario, nuplan_center):
    data = [
        parse_ego_vehicle_state(scenario.get_ego_state_at_iteration(i), nuplan_center)
        for i in range(scenario.get_number_of_iterations())
    ]
    for i in range(len(data) - 1):
        data[i]["angular_velocity"] = compute_angular_velocity(
            initial_heading=data[i]["heading"], final_heading=data[i + 1]["heading"], dt=scenario.database_interval
        )
    return data


def extract_traffic(scenario: NuPlanScenario, center):
    episode_len = scenario.get_number_of_iterations()
    detection_ret = []
    all_objs = set()
    all_objs.add(EGO)
    for frame_data in [scenario.get_tracked_objects_at_iteration(i).tracked_objects for i in range(episode_len)]:
        new_frame_data = {}
        for obj in frame_data:
            new_frame_data[obj.track_token] = obj
            all_objs.add(obj.track_token)
        detection_ret.append(new_frame_data)

    tracks = {
        k: dict(
            type=MetaDriveType.UNSET,
            state=dict(
                position=np.zeros(shape=(episode_len, 3)),
                heading=np.zeros(shape=(episode_len, )),
                velocity=np.zeros(shape=(episode_len, 2)),
                valid=np.zeros(shape=(episode_len, )),
                length=np.zeros(shape=(episode_len, 1)),
                width=np.zeros(shape=(episode_len, 1)),
                height=np.zeros(shape=(episode_len, 1))
            ),
            metadata=dict(track_length=episode_len, nuplan_type=None, type=None, object_id=k, nuplan_id=k)
        )
        for k in list(all_objs)
    }

    tracks_to_remove = set()

    for frame_idx, frame in enumerate(detection_ret):
        for nuplan_id, obj_state, in frame.items():
            assert isinstance(obj_state, Agent) or isinstance(obj_state, StaticObject)
            obj_type = get_traffic_obj_type(obj_state.tracked_object_type)
            if obj_type is None:
                tracks_to_remove.add(nuplan_id)
                continue
            tracks[nuplan_id][SD.TYPE] = obj_type
            if tracks[nuplan_id][SD.METADATA]["nuplan_type"] is None:
                tracks[nuplan_id][SD.METADATA]["nuplan_type"] = int(obj_state.tracked_object_type)
                tracks[nuplan_id][SD.METADATA]["type"] = obj_type

            state = parse_object_state(obj_state, center)
            tracks[nuplan_id]["state"]["position"][frame_idx] = [state["position"][0], state["position"][1], 0.0]
            tracks[nuplan_id]["state"]["heading"][frame_idx] = state["heading"]
            tracks[nuplan_id]["state"]["velocity"][frame_idx] = state["velocity"]
            tracks[nuplan_id]["state"]["valid"][frame_idx] = 1
            tracks[nuplan_id]["state"]["length"][frame_idx] = state["length"]
            tracks[nuplan_id]["state"]["width"][frame_idx] = state["width"]
            tracks[nuplan_id]["state"]["height"][frame_idx] = state["height"]

    for track in list(tracks_to_remove):
        tracks.pop(track)

    # ego
    sdc_traj = parse_ego_vehicle_state_trajectory(scenario, center)
    ego_track = tracks[EGO]

    for frame_idx, obj_state in enumerate(sdc_traj):
        obj_type = MetaDriveType.VEHICLE
        ego_track[SD.TYPE] = obj_type
        if ego_track[SD.METADATA]["nuplan_type"] is None:
            ego_track[SD.METADATA]["nuplan_type"] = int(TrackedObjectType.EGO)
            ego_track[SD.METADATA]["type"] = obj_type
        state = obj_state
        ego_track["state"]["position"][frame_idx] = [state["position"][0], state["position"][1], 0.0]
        ego_track["state"]["valid"][frame_idx] = 1
        ego_track["state"]["heading"][frame_idx] = state["heading"]
        # this velocity is in ego car frame, abort
        # ego_track["state"]["velocity"][frame_idx] = state["velocity"]

        ego_track["state"]["length"][frame_idx] = state["length"]
        ego_track["state"]["width"][frame_idx] = state["width"]
        ego_track["state"]["height"][frame_idx] = state["height"]

    # get velocity here
    vel = ego_track["state"]["position"][1:] - ego_track["state"]["position"][:-1]
    ego_track["state"]["velocity"][:-1] = vel[..., :2] / 0.1
    ego_track["state"]["velocity"][-1] = ego_track["state"]["velocity"][-2]

    # check
    assert EGO in tracks
    for track_id in tracks:
        assert tracks[track_id][SD.TYPE] != MetaDriveType.UNSET

    return tracks


def convert_one_scenario(scenario: NuPlanScenario):
    """
    Data will be interpolated to 0.1s time interval, while the time interval of original key frames are 0.5s.
    """
    scenario_log_interval = scenario.database_interval
    assert abs(scenario_log_interval - 0.1) < 1e-3, "Log interval should be 0.1 or Interpolating is required! " \
                                                    "By setting NuPlan subsample ratio can address this"

    result = SD()
    result[SD.ID] = scenario.scenario_name
    result[SD.VERSION] = "nuplan" + scenario.map_version
    result[SD.LENGTH] = scenario.get_number_of_iterations()
    # metadata
    result[SD.METADATA] = {}
    result[SD.METADATA]["dataset"] = "nuplan"
    result[SD.METADATA]["map"] = scenario.map_api.map_name
    result[SD.METADATA][SD.METADRIVE_PROCESSED] = False
    result[SD.METADATA]["map_version"] = scenario.map_version
    result[SD.METADATA]["log_name"] = scenario.log_name
    result[SD.METADATA]["scenario_extraction_info"] = scenario._scenario_extraction_info.__dict__
    result[SD.METADATA]["ego_vehicle_parameters"] = scenario.ego_vehicle_parameters.__dict__
    result[SD.METADATA]["coordinate"] = "right-handed"
    result[SD.METADATA]["scenario_token"] = scenario.scenario_name
    result[SD.METADATA]["scenario_id"] = scenario.scenario_name
    result[SD.METADATA]["scenario_type"] = scenario.scenario_type
    result[SD.METADATA]["sample_rate"] = scenario_log_interval
    result[SD.METADATA][SD.TIMESTEP] = np.asarray([i * scenario_log_interval for i in range(result[SD.LENGTH])])

    # centered all positions to ego car
    state = scenario.get_ego_state_at_iteration(0)
    scenario_center = [state.waypoint.x, state.waypoint.y]

    result[SD.TRACKS] = extract_traffic(scenario, scenario_center)
    result[SD.METADATA][SD.SDC_ID] = EGO

    # traffic light
    result[SD.DYNAMIC_MAP_STATES] = extract_traffic_light(scenario, scenario_center)

    # map
    result[SD.MAP_FEATURES] = extract_map_features(scenario.map_api, scenario_center)

    return result



from metadrive.component.block.base_block import BaseBlock
from metadrive.component.lane.opendrive_lane import OpenDriveLane
from metadrive.component.road_network.edge_road_network import OpenDriveRoadNetwork
from metadrive.utils.opendrive.map_load import get_lane_width


class OpenDriveBlock(BaseBlock):
    """
    The OpenDriveBlock instance will wrap a section in a Road, which serves as a basic element for building a map
    """
    def __init__(self, block_index: int, global_network, random_seed, section_data):
        self.section_data = section_data
        super(OpenDriveBlock, self).__init__(block_index, global_network, random_seed)

    def _sample_topology(self) -> bool:
        for lane in self.section_data.allLanes:
            # if lane.type == "driving":
            width = get_lane_width(lane)
            opendrive_lane = OpenDriveLane(width, lane)
            self.block_network.add_lane(opendrive_lane)
        return True

    def create_in_world(self):
        """
        The lane line should be created separately
        """
        graph = self.block_network.graph
        for id, lane_info in graph.items():
            lane = lane_info.lane
            lane.construct_lane_in_block(self, lane_index=id)

    @property
    def block_network_type(self):
        return OpenDriveRoadNetwork

    def destroy(self):
        self.section_data = None
        super(OpenDriveBlock, self).destroy()



from metadrive.utils.coordinates_shift import nuplan_to_metadrive_heading, nuplan_to_metadrive_vector
from metadrive.utils.math import compute_angular_velocity


def parse_ego_vehicle_state(state, nuplan_center):
    center = nuplan_center
    ret = {}
    ret["position"] = nuplan_to_metadrive_vector([state.waypoint.x, state.waypoint.y], center)
    ret["heading"] = nuplan_to_metadrive_heading(state.waypoint.heading)
    ret["velocity"] = nuplan_to_metadrive_vector([state.agent.velocity.x, state.agent.velocity.y])
    ret["angular_velocity"] = nuplan_to_metadrive_heading(state.dynamic_car_state.angular_velocity)
    ret["valid"] = True
    ret["length"] = state.agent.box.length
    ret["width"] = state.agent.box.width
    return ret


def parse_ego_vehicle_state_trajectory(scenario, nuplan_center):
    data = [
        parse_ego_vehicle_state(scenario.get_ego_state_at_iteration(i), nuplan_center)
        for i in range(scenario.get_number_of_iterations())
    ]
    for i in range(len(data) - 1):
        data[i]["angular_velocity"] = compute_angular_velocity(
            initial_heading=data[i]["heading"], final_heading=data[i + 1]["heading"], dt=scenario.database_interval
        )
    return data


def parse_ego_vehicle_trajectory(states, nuplan_center):
    traj = []
    for state in states:
        traj.append([state.waypoint.x, state.waypoint.y])
    trajectory = nuplan_to_metadrive_vector(traj, nuplan_center=nuplan_center)
    return trajectory


def parse_object_state(obj_state, nuplan_center):
    ret = {}
    ret["position"] = nuplan_to_metadrive_vector([obj_state.center.x, obj_state.center.y], nuplan_center)
    ret["heading"] = nuplan_to_metadrive_heading(obj_state.center.heading)
    ret["velocity"] = nuplan_to_metadrive_vector([obj_state.velocity.x, obj_state.velocity.y])
    ret["valid"] = True
    ret["length"] = obj_state.box.length
    ret["width"] = obj_state.box.width
    return ret



import unittest

from metadrive import MetaDriveEnv


class TestObsActionSpace(unittest.TestCase):
    def setUp(self):
        self.env = MetaDriveEnv()

    def test_obs_space(self):
        obs, _ = self.env.reset()
        assert self.env.observation_space.contains(obs), (self.env.observation_space, obs.shape)
        obs, _, _, _, _ = self.env.step(self.env.action_space.sample())
        assert self.env.observation_space.contains(obs), (self.env.observation_space, obs.shape)

    def tearDown(self):
        self.env.close()


if __name__ == '__main__':
    unittest.main()



"""

RenderPipeline

Copyright (c) 2014-2016 tobspr <tobias.springer1@gmail.com>

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
THE SOFTWARE.

"""

from __future__ import division
from metadrive.render_pipeline.rplibs.six.moves import range  # pylint: disable=import-error
from metadrive.render_pipeline.rplibs.six import itervalues

from panda3d.core import Camera, PerspectiveLens, Vec4, Vec3, PTAInt

from metadrive.render_pipeline.rpcore.globals import Globals
from metadrive.render_pipeline.rpcore.image import Image
from metadrive.render_pipeline.rpcore.render_stage import RenderStage


class EnvironmentCaptureStage(RenderStage):
    """ This stage renders the scene to a cubemap """

    required_inputs = ["DefaultEnvmap", "AllLightsData", "maxLightIndex", "IESDatasetTex"]
    required_pipes = []

    def __init__(self, pipeline):
        RenderStage.__init__(self, pipeline)
        self.resolution = 128
        self.diffuse_resolution = 4
        self.regions = []
        self.cameras = []
        self.rig_node = Globals.render.attach_new_node("EnvmapCamRig")
        self.pta_index = PTAInt.empty_array(1)
        self.storage_tex = None
        self.storage_tex_diffuse = None

    def create(self):
        self.target = self.create_target("CaptureScene")
        self.target.size = self.resolution * 6, self.resolution
        self.target.add_depth_attachment(bits=16)
        self.target.add_color_attachment(bits=16, alpha=True)
        self.target.prepare_render(None)

        # Remove all unused display regions
        internal_buffer = self.target.internal_buffer
        internal_buffer.remove_all_display_regions()
        internal_buffer.disable_clears()
        internal_buffer.get_overlay_display_region().disable_clears()

        self._setup_camera_rig()
        self._create_store_targets()
        self._create_filter_targets()

    def _setup_camera_rig(self):
        """ Setups the cameras to render a cubemap """
        directions = (Vec3(1, 0, 0), Vec3(-1, 0, 0), Vec3(0, 1, 0), Vec3(0, -1, 0), Vec3(0, 0, 1), Vec3(0, 0, -1))

        # Prepare the display regions
        for i in range(6):
            region = self.target.internal_buffer.make_display_region(i / 6, i / 6 + 1 / 6, 0, 1)
            region.set_sort(25 + i)
            region.set_active(True)
            region.disable_clears()

            # Set the correct clears
            region.set_clear_depth_active(True)
            region.set_clear_depth(1.0)
            region.set_clear_color_active(True)
            region.set_clear_color(Vec4(0))

            lens = PerspectiveLens()
            lens.set_fov(90)
            lens.set_near_far(0.001, 1.0)
            camera = Camera("EnvmapCam-" + str(i), lens)
            camera_np = self.rig_node.attach_new_node(camera)
            camera_np.look_at(camera_np, directions[i])
            region.set_camera(camera_np)
            self.regions.append(region)
            self.cameras.append(camera_np)

        self.cameras[0].set_r(90)
        self.cameras[1].set_r(-90)
        self.cameras[3].set_r(180)
        self.cameras[5].set_r(180)

        # Register cameras
        for camera_np in self.cameras:
            self._pipeline.tag_mgr.register_camera("envmap", camera_np.node())

    def _create_store_targets(self):
        """ Creates the targets which copy the result texture into the actual storage """
        self.target_store = self.create_target("StoreCubemap")
        self.target_store.size = self.resolution * 6, self.resolution
        self.target_store.prepare_buffer()
        self.target_store.set_shader_inputs(
            SourceTex=self.target.color_tex, DestTex=self.storage_tex, currentIndex=self.pta_index
        )

        self.temporary_diffuse_map = Image.create_cube("DiffuseTemp", self.resolution, "RGBA16")
        self.target_store_diff = self.create_target("StoreCubemapDiffuse")
        self.target_store_diff.size = self.resolution * 6, self.resolution
        self.target_store_diff.prepare_buffer()
        self.target_store_diff.set_shader_inputs(
            SourceTex=self.target.color_tex, DestTex=self.temporary_diffuse_map, currentIndex=self.pta_index
        )

    def _create_filter_targets(self):
        """ Generates the targets which filter the specular cubemap """
        self.filter_targets = []
        mip = 0
        size = self.resolution
        while size > 1:
            size = size // 2
            mip += 1
            target = self.create_target("FilterCubemap:{0}-{1}x{1}".format(mip, size))
            target.size = size * 6, size
            target.prepare_buffer()
            target.set_shader_inputs(currentIndex=self.pta_index, currentMip=mip, SourceTex=self.storage_tex)
            target.set_shader_input("DestTex", self.storage_tex, False, True, -1, mip, 0)
            self.filter_targets.append(target)

        # Target to filter the diffuse cubemap
        self.filter_diffuse_target = self.create_target("FilterCubemapDiffuse")
        self.filter_diffuse_target.size = self.diffuse_resolution * 6, self.diffuse_resolution
        self.filter_diffuse_target.prepare_buffer()
        self.filter_diffuse_target.set_shader_inputs(
            SourceTex=self.temporary_diffuse_map, DestTex=self.storage_tex_diffuse, currentIndex=self.pta_index
        )

    def set_probe(self, probe):
        self.rig_node.set_mat(probe.matrix)
        self.pta_index[0] = probe.index

    def update(self):
        # First, disable all targets
        for target in itervalues(self._targets):
            target.active = False

        # Check for updated faces
        for i in range(6):
            if self._pipeline.task_scheduler.is_scheduled("envprobes_capture_envmap_face" + str(i)):
                self.regions[i].set_active(True)

        # Check for filtering
        if self._pipeline.task_scheduler.is_scheduled("envprobes_filter_and_store_envmap"):
            self.target_store.active = True
            self.target_store_diff.active = True
            self.filter_diffuse_target.active = True
            for target in self.filter_targets:
                target.active = True

    def set_shader_input(self, *args):
        Globals.render.set_shader_input(*args)

    def set_shader_inputs(self, **kwargs):
        Globals.render.set_shader_inputs(**kwargs)

    def reload_shaders(self):
        self.target_store.shader = self.load_plugin_shader("store_cubemap.frag.glsl")
        self.target_store_diff.shader = self.load_plugin_shader("store_cubemap_diffuse.frag.glsl")
        self.filter_diffuse_target.shader = self.load_plugin_shader("filter_cubemap_diffuse.frag.glsl")
        for i, target in enumerate(self.filter_targets):
            target.shader = self.load_plugin_shader("mips/{}.autogen.glsl".format(i))
